# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Miscellaneous helpers."""
import builtins
import contextlib
import datetime
import importlib.util
import logging
import os
import platform
import re
import shutil
import signal
import subprocess  # nosec
import sys
import time
import types
from collections import OrderedDict, UserString, defaultdict, deque
from collections.abc import Mapping
from copy import copy
from functools import wraps
from importlib.machinery import ModuleSpec
from pathlib import Path
from threading import RLock
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from dotenv import load_dotenv
from packaging.version import Version

from aea.common import PathLike
from aea.exceptions import enforce


STRING_LENGTH_LIMIT = 128
ISO_8601_DATE_FORMAT = "%Y-%m-%d"

_default_logger = logging.getLogger(__name__)


def _get_module(spec: ModuleSpec) -> Optional[types.ModuleType]:
    """Try to execute a module. Return None if the attempt fail."""
    try:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore
        return module
    except Exception:  # pylint: disable=broad-except
        return None


def locate(path: str) -> Any:
    """Locate an object by name or dotted save_path, importing as necessary."""
    parts = [part for part in path.split(".") if part]
    module, n = None, 0
    while n < len(parts):
        file_location = os.path.join(*parts[: n + 1])
        spec_name = ".".join(parts[: n + 1])
        module_location = os.path.join(file_location, "__init__.py")
        spec = importlib.util.spec_from_file_location(spec_name, module_location)
        _default_logger.debug("Trying to import {}".format(module_location))
        nextmodule = _get_module(spec)
        if nextmodule is None:
            module_location = file_location + ".py"
            spec = importlib.util.spec_from_file_location(spec_name, module_location)
            _default_logger.debug("Trying to import {}".format(module_location))
            nextmodule = _get_module(spec)

        if nextmodule:
            module, n = nextmodule, n + 1
        else:  # pragma: nocover
            break
    if module:
        object_ = module
    else:
        object_ = builtins
    for part in parts[n:]:
        try:
            object_ = getattr(object_, part)
        except AttributeError:
            return None
    return object_


def load_module(dotted_path: str, filepath: Path) -> types.ModuleType:
    """
    Load a module.

    :param dotted_path: the dotted save_path of the package/module.
    :param filepath: the file to the package/module.
    :return: None
    :raises ValueError: if the filepath provided is not a module.
    :raises Exception: if the execution of the module raises exception.
    """
    spec = importlib.util.spec_from_file_location(dotted_path, str(filepath))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)  # type: ignore
    return module


def load_env_file(env_file: str) -> None:
    """
    Load the content of the environment file into the process environment.

    :param env_file: save_path to the env file.
    :return: None.
    """
    load_dotenv(dotenv_path=Path(env_file), override=False)


def sigint_crossplatform(process: subprocess.Popen) -> None:  # pragma: nocover
    """
    Send a SIGINT, cross-platform.

    The reason is because the subprocess module
    doesn't have an API to send a SIGINT-like signal
    both on Posix and Windows with a single method.

    However, a subprocess.Popen class has the method
    'send_signal' that gives more flexibility in this terms.

    :param process: the process to send the signal to.
    :return: None
    """
    if os.name == "posix":
        process.send_signal(signal.SIGINT)  # pylint: disable=no-member
    elif os.name == "nt":
        process.send_signal(signal.CTRL_C_EVENT)  # pylint: disable=no-member
    else:
        raise ValueError("Other platforms not supported.")


def win_popen_kwargs() -> dict:
    """
    Return kwargs to start a process in windows with new process group.

    Help to handle ctrl c properly.
    Return empty dict if platform is not win32
    """
    kwargs: dict = {}

    if sys.platform == "win32":  # pragma: nocover
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        kwargs["startupinfo"] = startupinfo
        kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP

    return kwargs


def send_control_c(
    process: subprocess.Popen, kill_group: bool = False
) -> None:  # pragma: nocover # cause platform dependent
    """
    Send ctrl-C crossplatform to terminate a subprocess.

    :param process: the process to send the signal to.

    :return: None
    """
    if platform.system() == "Windows":
        if process.stdin:  # cause ctrl-c event will be handled with stdin
            process.stdin.close()
        os.kill(process.pid, signal.CTRL_C_EVENT)  # pylint: disable=no-member
    elif kill_group:
        pgid = os.getpgid(process.pid)
        os.killpg(pgid, signal.SIGINT)
    else:
        os.kill(process.pid, signal.SIGINT)


class RegexConstrainedString(UserString):
    """
    A string that is constrained by a regex.

    The default behaviour is to match anything.
    Subclass this class and change the 'REGEX' class
    attribute to implement a different behaviour.
    """

    REGEX = re.compile(".*", flags=re.DOTALL)

    def __init__(self, seq: Union[UserString, str]) -> None:
        """Initialize a regex constrained string."""
        super().__init__(seq)

        if not self.REGEX.match(self.data):
            self._handle_no_match()

    def _handle_no_match(self) -> None:
        raise ValueError(
            "Value {data} does not match the regular expression {regex}".format(
                data=self.data, regex=self.REGEX
            )
        )


class SimpleId(RegexConstrainedString):
    """
    A simple identifier.

    The allowed strings are all the strings that:
    - have at least length 1
    - have at most length 128
    - the first character must be between a-z,A-Z or underscore
    - the other characters must be either the above or digits.

    Examples of allowed strings:
    >>> SimpleId("an_identifier")
    'an_identifier'

    Examples of not allowed strings:
    >>> SimpleId("0an_identifier")
    Traceback (most recent call last):
    ...
    ValueError: Value 0an_identifier does not match the regular expression re.compile('[a-zA-Z_][a-zA-Z0-9_]{0,127}')

    >>> SimpleId("")
    Traceback (most recent call last):
    ...
    ValueError: Value  does not match the regular expression re.compile('[a-zA-Z_][a-zA-Z0-9_]{0,127}')
    """

    REGEX = re.compile(fr"[a-zA-Z_][a-zA-Z0-9_]{{0,{STRING_LENGTH_LIMIT - 1}}}")


SimpleIdOrStr = Union[SimpleId, str]


@contextlib.contextmanager
def cd(path: PathLike) -> Generator:  # pragma: nocover
    """Change working directory temporarily."""
    old_path = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_path)


def get_logger_method(fn: Callable, logger_method: Union[str, Callable]) -> Callable:
    """
    Get logger method for function.

    Get logger in `fn` definion module or creates logger is module.__name__.
    Or return logger_method if it's callable.

    :param fn: function to get logger for.
    :param logger_method: logger name or callable.

    :return: callable to write log with
    """
    if callable(logger_method):  # pragma: nocover
        return logger_method

    logger_ = fn.__globals__.get("logger", logging.getLogger(fn.__globals__["__name__"]))  # type: ignore

    return getattr(logger_, logger_method)


def try_decorator(
    error_message: str, default_return: Callable = None, logger_method: Any = "error"
) -> Callable:
    """
    Run function, log and return default value on exception.

    Does not support async or coroutines!

    :param error_message: message template with one `{}` for exception
    :param default_return: value to return on exception, by default None
    :param logger_method: name of the logger method or callable to print logs
    """

    # for pydocstyle
    def decorator(fn: Callable) -> Callable:
        @wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Callable:
            try:
                return fn(*args, **kwargs)
            except Exception as e:  # pylint: disable=broad-except  # pragma: no cover  # generic code
                if error_message:
                    log = get_logger_method(fn, logger_method)
                    log(error_message.format(e))
                return cast(Callable, default_return)

        return wrapper

    return decorator


class MaxRetriesError(Exception):
    """Exception for retry decorator."""


def retry_decorator(
    number_of_retries: int,
    error_message: str,
    delay: float = 0,
    logger_method: str = "error",
) -> Callable:
    """
    Run function with several attempts.

    Does not support async or coroutines!

    :param number_of_retries: amount of attempts
    :param error_message: message template with one `{}` for exception
    :param delay: num of seconds to sleep between retries. default 0
    :param logger_method: name of the logger method or callable to print logs
    """

    # for pydocstyle
    def decorator(fn: Callable) -> Callable:
        @wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Callable:
            log = get_logger_method(fn, logger_method)
            for retry in range(number_of_retries):
                try:
                    return fn(*args, **kwargs)
                except Exception as e:  # pylint: disable=broad-except  # pragma: no cover  # generic code
                    if error_message:
                        log(error_message.format(retry=retry + 1, error=e))
                    if delay:
                        time.sleep(delay)
            raise MaxRetriesError(number_of_retries)

        return wrapper

    return decorator


@contextlib.contextmanager
def exception_log_and_reraise(log_method: Callable, message: str) -> Generator:
    """
    Run code in context to log and re raise exception.

    :param log_method: function to print log
    :param message: message template to add error text.
    """
    try:
        yield
    except BaseException as e:  # pylint: disable=broad-except  # pragma: no cover  # generic code
        log_method(message.format(e))
        raise


def _is_dict_like(obj: Any) -> bool:
    """
    Check whether an object is dict-like (i.e. either dict or OrderedDict).

    :param obj: the object to test.
    :return: True if the object is dict-like, False otherwise.
    """
    return type(obj) in {dict, OrderedDict}


def recursive_update(
    to_update: Dict, new_values: Dict, allow_new_values: bool = False,
) -> None:
    """
    Update a dictionary by replacing conflicts with the new values.

    It does side-effects to the first dictionary.

    >>> to_update = dict(a=1, b=2, subdict=dict(subfield1=1))
    >>> new_values = dict(b=3, subdict=dict(subfield1=2))
    >>> recursive_update(to_update, new_values)
    >>> to_update
    {'a': 1, 'b': 3, 'subdict': {'subfield1': 2}}

    :param to_update: the dictionary to update.
    :param new_values: the dictionary of new values to replace.
    :return: None
    """
    for key, value in new_values.items():
        if (not allow_new_values) and key not in to_update:
            raise ValueError(
                f"Key '{key}' is not contained in the dictionary to update."
            )

        if key not in to_update and allow_new_values:
            to_update[key] = value
            continue

        value_to_update = to_update[key]
        value_type = type(value)
        value_to_update_type = type(value_to_update)
        both_are_dict = _is_dict_like(value) and _is_dict_like(value_to_update)
        if (
            not both_are_dict
            and value_type != value_to_update_type
            and value is not None
            and value_to_update is not None
        ):
            raise ValueError(
                f"Trying to replace value '{value_to_update}' with value '{value}' which is of different type."
            )

        if both_are_dict:
            recursive_update(value_to_update, value, allow_new_values)
        else:
            to_update[key] = value


def _get_aea_logger_name_prefix(module_name: str, agent_name: str) -> str:
    """
    Get the logger name prefix.

    It consists of a dotted save_path with:
    - the name of the package, 'aea';
    - the agent name;
    - the rest of the dotted save_path.

    >>> _get_aea_logger_name_prefix("aea.save_path.to.package", "myagent")
    'aea.myagent.save_path.to.package'

    :param module_name: the module name.
    :param agent_name: the agent name.
    :return: the logger name prefix.
    """
    module_name_parts = module_name.split(".")
    root = module_name_parts[0]
    postfix = module_name_parts[1:]
    return ".".join([root, agent_name, *postfix])


T = TypeVar("T")


def find_topological_order(adjacency_list: Dict[T, Set[T]]) -> List[T]:
    """
    Compute the topological order of a graph (using Kahn's algorithm).

    :param adjacency_list: the adjacency list of the graph.
    :return: the topological order for the graph (as a sequence of nodes)
    :raises ValueError: if the graph contains a cycle.
    """
    # compute inverse adjacency list and the roots of the DAG.
    adjacency_list = copy(adjacency_list)
    visited: Set[T] = set()
    roots: Set[T] = set()
    inverse_adjacency_list: Dict[T, Set[T]] = defaultdict(set)
    # compute both roots and inv. adj. list in one pass.
    for start_node, end_nodes in adjacency_list.items():
        if start_node not in visited:
            roots.add(start_node)
        visited.update([start_node, *end_nodes])
        for end_node in end_nodes:
            roots.discard(end_node)
            inverse_adjacency_list[end_node].add(start_node)

    # compute the topological order
    queue: Deque[T] = deque()
    order = []
    queue.extendleft(sorted(roots))
    while len(queue) > 0:
        current = queue.pop()
        order.append(current)
        next_nodes = adjacency_list.get(current, set())
        for node in next_nodes:
            inverse_adjacency_list[node].discard(current)
            if len(inverse_adjacency_list[node]) == 0:
                queue.append(node)

        # remove all the edges
        adjacency_list[current] = set()

    if any(len(edges) > 0 for edges in inverse_adjacency_list.values()):
        raise ValueError("Graph has at least one cycle.")

    return order


def reachable_nodes(
    adjacency_list: Dict[T, Set[T]], starting_nodes: Set[T]
) -> Dict[T, Set[T]]:
    """
    Find the reachable subgraph induced by a set of starting nodes.

    :param adjacency_list: the adjacency list of the full graph.
    :param starting_nodes: the starting nodes of the new graph.
    :return: the adjacency list of the subgraph.
    """
    all_nodes = set()
    for node, nodes in adjacency_list.items():
        all_nodes.add(node)
        all_nodes.update(nodes)
    enforce(
        all(s in all_nodes for s in starting_nodes),
        f"These starting nodes are not in the set of nodes: {starting_nodes.difference(all_nodes)}",
    )
    visited: Set[T] = set()
    result: Dict[T, Set[T]] = {start_node: set() for start_node in starting_nodes}
    queue: Deque[T] = deque()
    queue.extend(starting_nodes)
    while len(queue) > 0:
        current = queue.pop()
        if current in visited or current not in adjacency_list:
            continue
        successors = adjacency_list.get(current, set())
        result.setdefault(current, set()).update(successors)
        queue.extendleft(successors)
        visited.add(current)
    return result


_NOT_FOUND = object()


# copied from python3.8 functools
class cached_property:  # pragma: nocover
    """Cached property from python3.8 functools."""

    def __init__(self, func: Callable) -> None:
        """Init cached property."""
        self.func = func
        self.attrname = None
        self.__doc__ = func.__doc__
        self.lock = RLock()

    def __set_name__(self, _: Any, name: Any) -> None:
        """Set name."""
        if self.attrname is None:
            self.attrname = name
        elif name != self.attrname:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                f"({self.attrname!r} and {name!r})."
            )

    def __get__(self, instance: Any, _: Optional[Any] = None) -> Any:
        """Get instance."""
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError(
                "Cannot use cached_property instance without calling __set_name__ on it."
            )
        try:
            cache = instance.__dict__
        except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {self.attrname!r} property."
            )
            raise TypeError(msg) from None
        val = cache.get(self.attrname, _NOT_FOUND)
        if val is _NOT_FOUND:
            with self.lock:
                # check if another thread filled cache while we awaited lock
                val = cache.get(self.attrname, _NOT_FOUND)
                if val is _NOT_FOUND:
                    val = self.func(instance)
                    try:
                        cache[self.attrname] = val
                    except TypeError:
                        msg = (
                            f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                            f"does not support item assignment for caching {self.attrname!r} property."
                        )
                        raise TypeError(msg) from None
        return val


def ensure_dir(dir_path: str) -> None:
    """Check if dir_path is a directory or create it."""
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    else:
        enforce(os.path.isdir(dir_path), f"{dir_path} is not a directory!")


def dict_to_path_value(
    data: Mapping, path: Optional[List] = None
) -> Iterable[Tuple[List[str], Any]]:
    """Convert dict to sequence of terminal path build of  keys and value."""
    path = path or []
    for key, value in data.items():
        if isinstance(value, Mapping) and value:
            # terminal value
            for p, v in dict_to_path_value(value, path + [key]):
                yield p, v
        else:
            yield path + [key], value


def parse_datetime_from_str(date_string: str) -> datetime.datetime:
    """Parse datetime from string."""
    result = datetime.datetime.strptime(date_string, ISO_8601_DATE_FORMAT)
    result = result.replace(tzinfo=datetime.timezone.utc)
    return result


class CertRequest:
    """Certificate request for proof of representation."""

    def __init__(
        self,
        public_key: str,
        identifier: SimpleIdOrStr,
        ledger_id: SimpleIdOrStr,
        not_before: str,
        not_after: str,
        save_path: str,
    ) -> None:
        """
        Initialize the certificate request.

        :param public_key: the public key, or the key id.
        :param identifier: certificate identifier.
        :param not_before: specify the lower bound for certificate validity.
          If it is a string, it must follow the format: 'YYYY-MM-DD'. It
          will be interpreted as timezone UTC.
        :param not_before: specify the lower bound for certificate validity.
          if it is a string, it must follow the format: 'YYYY-MM-DD' It
          will be interpreted as timezone UTC-0.
        :param save_path: the save_path where to save the certificate.
        """
        self._key_identifier: Optional[str] = None
        self._public_key: Optional[str] = None
        self._identifier = str(SimpleId(identifier))
        self._ledger_id = str(SimpleId(ledger_id))
        self._not_before_string = not_before
        self._not_after_string = not_after
        self._not_before = self._parse_datetime(not_before)
        self._not_after = self._parse_datetime(not_after)
        self._save_path = Path(save_path)

        self._parse_public_key(public_key)
        self._check_validation_boundaries()

    @classmethod
    def _parse_datetime(cls, obj: Union[str, datetime.datetime]) -> datetime.datetime:
        """
        Parse datetime string.

        It is expected to follow ISO 8601.

        :param obj: the input to parse.
        :return: a datetime.datetime instance.
        """
        result = (
            parse_datetime_from_str(obj)  # type: ignore
            if isinstance(obj, str)
            else obj
        )
        enforce(result.microsecond == 0, "Microsecond field not allowed.")
        return result

    def _check_validation_boundaries(self) -> None:
        """
        Check the validation boundaries are consistent.

        Namely, that not_before < not_after.
        """
        enforce(
            self._not_before < self._not_after,
            f"Inconsistent certificate validity period: 'not_before' field '{self._not_before_string}' is not before than 'not_after' field '{self._not_after_string}'",
            ValueError,
        )

    def _parse_public_key(self, public_key_str: str) -> None:
        """
        Parse public key from string.

        It first tries to parse it as an identifier,
        and in case of failure as a sequence of hexadecimals, starting with "0x".
        """
        with contextlib.suppress(ValueError):
            # if this raises ValueError, we don't return
            self._key_identifier = str(SimpleId(public_key_str))
            return

        with contextlib.suppress(ValueError):
            # this raises ValueError if the input is not a valid hexadecimal string.
            int(public_key_str, 16)
            self._public_key = public_key_str
            return

        enforce(
            False,
            f"Public key field '{public_key_str}' is neither a valid identifier nor an address.",
            exception_class=ValueError,
        )

    @property
    def public_key(self) -> Optional[str]:
        """Get the public key."""
        return self._public_key

    @property
    def ledger_id(self) -> str:
        """Get the ledger id."""
        return self._ledger_id

    @property
    def key_identifier(self) -> Optional[str]:
        """Get the key identifier."""
        return self._key_identifier

    @property
    def identifier(self) -> str:
        """Get the identifier."""
        return self._identifier

    @property
    def not_before_string(self) -> str:
        """Get the not_before field as string."""
        return self._not_before_string

    @property
    def not_after_string(self) -> str:
        """Get the not_after field as string."""
        return self._not_after_string

    @property
    def not_before(self) -> datetime.datetime:
        """Get the not_before field."""
        return self._not_before

    @property
    def not_after(self) -> datetime.datetime:
        """Get the not_after field."""
        return self._not_after

    @property
    def save_path(self) -> Path:
        """
        Get the save path for the certificate.

        Note: if the path is *not* absolute, then
        the actual save path might depend on the context.
        """
        return self._save_path

    def get_absolute_save_path(self, path_prefix: Optional[PathLike] = None) -> Path:
        """
        Get the absolute save path.

        If save_path is an absolute path, then the prefix is ignored.
        Otherwise, the path prefix is prepended.

        :param path_prefix: the (absolute) path to prepend to the save path.
        :return: the actual save path.
        """
        path_prefix = (
            Path(path_prefix).absolute() if path_prefix is not None else Path.cwd()
        )
        return (
            self.save_path
            if self.save_path.is_absolute()
            else path_prefix / self.save_path
        )

    @property
    def public_key_or_identifier(self) -> str:
        """Get the public key or identifier."""
        if (self.public_key is None and self.key_identifier is None) or (
            self.public_key is not None and self.key_identifier is not None
        ):
            raise ValueError(  # pragma: nocover
                "Exactly one of key_identifier or public_key can be specified."
            )
        if self.public_key is not None:
            result = self.public_key
        elif self.key_identifier is not None:
            result = self.key_identifier
        return result

    def get_message(self, public_key: str) -> bytes:  # pylint: disable=no-self-use
        """Get the message to sign."""
        message = public_key.encode("ascii")
        # + self.identifier.encode("ascii")  # noqa: E800
        # + self.not_before_string.encode("ascii")  # noqa: E800
        # + self.not_after_string.encode("ascii")  # noqa: E800
        return message

    def get_signature(self, path_prefix: Optional[PathLike] = None) -> str:
        """
        Get signature from save_path.

        :param path_prefix: the path prefix to be prependend to save_path. Defaults to cwd.
        :return: the signature.
        """
        save_path = self.get_absolute_save_path(path_prefix)
        if not Path(save_path).is_file():
            raise Exception(  # pragma: no cover
                f"cert_request 'save_path' field {save_path} is not a file. "
                "Please ensure that 'issue-certificates' command is called beforehand."
            )
        signature = bytes.fromhex(Path(save_path).read_bytes().decode("ascii")).decode(
            "ascii"
        )
        return signature

    @property
    def json(self) -> Dict:
        """Compute the JSON representation."""
        result = dict(
            identifier=self.identifier,
            ledger_id=self.ledger_id,
            not_before=self._not_before_string,
            not_after=self._not_after_string,
            public_key=self.public_key_or_identifier,
            save_path=str(self.save_path),
        )
        return result

    @classmethod
    def from_json(cls, obj: Dict) -> "CertRequest":
        """Compute the JSON representation."""
        return cls(**obj)

    def __eq__(self, other: Any) -> bool:
        """Check equality."""
        return (
            isinstance(other, CertRequest)
            and self.identifier == other.identifier
            and self.ledger_id == other.ledger_id
            and self.public_key == other.public_key
            and self.key_identifier == other.key_identifier
            and self.not_after == other.not_after
            and self.not_before == other.not_before
            and self.save_path == other.save_path
        )


def compute_specifier_from_version(version: Version) -> str:
    """
    Compute the specifier set from a version, by varying only on the patch number.

    I.e. from "{major}.{minor}.{patch}", return

    ">={major}.{minor}.0, <{major}.{minor + 1}.0"

    :param version: the version
    :return: the specifier set
    """
    new_major = version.major
    new_minor_low = version.minor
    new_minor_high = new_minor_low + 1
    lower_bound = Version(f"{new_major}.{new_minor_low}.0")
    upper_bound = Version(f"{new_major}.{new_minor_high}.0")
    specifier_set = f">={lower_bound}, <{upper_bound}"
    return specifier_set


def decorator_with_optional_params(decorator: Callable) -> Callable:
    """
    Make a decorator usable either with or without parameters.

    In other words, if a decorator "mydecorator" is decorated with this decorator,
    It can be used both as:

    @mydecorator
    def myfunction():
        ...

    or as:

    @mydecorator(arg1, kwarg1="value")
    def myfunction():
        ...

    """

    @wraps(decorator)
    def new_decorator(*args: Any, **kwargs: Any) -> Callable:
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return decorator(args[0])

        def final_decorator(real_function: Callable) -> Callable:
            return decorator(real_function, *args, **kwargs)

        return final_decorator

    return new_decorator


def delete_directory_contents(directory: Path) -> None:
    """Delete the content of a directory, without deleting it."""
    enforce(directory.is_dir(), f"Path '{directory}' must be a directory.")
    for filename in directory.iterdir():
        if filename.is_file() or filename.is_symlink():
            filename.unlink()
        elif filename.is_dir():
            shutil.rmtree(str(filename), ignore_errors=False)


def prepend_if_not_absolute(path: PathLike, prefix: PathLike) -> PathLike:
    """
    Prepend a path with a prefix, but only if not absolute

    :param path: the path to process.
    :param prefix: the path prefix.
    :return: the same path if absolute, else the prepended path.
    """
    return path if Path(path).is_absolute() else Path(prefix) / path



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains Serializers that can be used for custom types."""

import copy
from typing import Any, Dict

from google.protobuf.struct_pb2 import Struct


class DictProtobufStructSerializer:
    """
    Serialize python dictionaries of type DictType = Dict[str, ValueType] recursively conserving their dynamic type, using google.protobuf.Struct

    ValueType = PrimitiveType | DictType | List[ValueType]]
    PrimitiveType = bool | int | float | str | bytes
    """

    NEED_PATCH = "_need_patch"

    @classmethod
    def encode(cls, dictionary: Dict[str, Any]) -> bytes:
        """Serialize compatible dictionary to bytes"""
        if not isinstance(dictionary, dict):
            raise TypeError(  # pragma: nocover
                "dictionary must be of dict type, got type {}".format(type(dictionary))
            )
        # TOFIX(LR) problematic as it'll copy every message
        patched_dict = copy.deepcopy(dictionary)
        cls._patch_dict(patched_dict)
        pstruct = Struct()
        pstruct.update(patched_dict)  # pylint: disable=no-member
        return pstruct.SerializeToString()

    @classmethod
    def decode(cls, buffer: bytes) -> Dict[str, Any]:
        """Deserialize a compatible dictionary"""
        pstruct = Struct()
        pstruct.ParseFromString(buffer)
        dictionary = dict(pstruct)
        cls._patch_dict_restore(dictionary)
        return dictionary

    @classmethod
    def _bytes_to_str(cls, value: bytes) -> str:
        return value.decode("utf-8")

    @classmethod
    def _str_to_bytes(cls, value: str) -> bytes:
        return value.encode("utf-8")

    @classmethod
    def _patch_dict(cls, dictionnary: Dict[str, Any]) -> None:
        need_patch: Dict[str, bool] = dict()
        for key, value in dictionnary.items():
            if isinstance(value, bytes):
                # convert bytes values to string, as protobuf.Struct does support byte fields
                dictionnary[key] = cls._bytes_to_str(value)
                if cls.NEED_PATCH in dictionnary:
                    dictionnary[cls.NEED_PATCH][key] = True
                else:
                    need_patch[key] = True
            elif isinstance(value, int) and not isinstance(value, bool):
                # protobuf Struct store int as float under numeric_value type
                if cls.NEED_PATCH in dictionnary:
                    dictionnary[cls.NEED_PATCH][key] = True
                else:
                    need_patch[key] = True
            elif isinstance(value, dict):
                cls._patch_dict(value)  # pylint: disable=protected-access
            elif (
                not isinstance(value, bool)
                and not isinstance(value, float)
                and not isinstance(value, str)
                and not isinstance(value, Struct)
            ):  # pragma: nocover
                raise NotImplementedError(
                    "DictProtobufStructSerializer doesn't support dict value type {}".format(
                        type(value)
                    )
                )
        if len(need_patch) > 0:
            dictionnary[cls.NEED_PATCH] = need_patch

    @classmethod
    def _patch_dict_restore(cls, dictionary: Dict[str, Any]) -> None:
        # protobuf Struct doesn't recursively convert Struct to dict
        need_patch = dictionary.get(cls.NEED_PATCH, {})
        if len(need_patch) > 0:
            dictionary[cls.NEED_PATCH] = dict(need_patch)

        for key, value in dictionary.items():
            if key == cls.NEED_PATCH:
                continue

            # protobuf struct doesn't recursively convert Struct to dict
            if isinstance(value, Struct):
                if value != Struct():
                    value = dict(value)
                dictionary[key] = value

            if isinstance(value, dict):
                cls._patch_dict_restore(value)
            elif isinstance(value, str) and dictionary.get(cls.NEED_PATCH, dict()).get(
                key, False
            ):
                dictionary[key] = cls._str_to_bytes(value)
            elif isinstance(value, float) and dictionary.get(
                cls.NEED_PATCH, dict()
            ).get(key, False):
                dictionary[key] = int(value)

        dictionary.pop(cls.NEED_PATCH, None)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Implementation of the 'aea issue_certificates' subcommand."""
from typing import Dict, List, Optional, cast

import click
from click import ClickException

from aea.cli.utils.context import Context
from aea.cli.utils.decorators import check_aea_project
from aea.cli.utils.loggers import logger
from aea.cli.utils.package_utils import get_dotted_package_path_unified
from aea.configurations.base import AgentConfig, PublicId
from aea.configurations.constants import CONNECTION
from aea.configurations.manager import AgentConfigManager, VariableDoesNotExist
from aea.crypto.helpers import make_certificate
from aea.crypto.registries import crypto_registry
from aea.exceptions import enforce
from aea.helpers.base import CertRequest, prepend_if_not_absolute


@click.command()
@click.pass_context
@check_aea_project
def issue_certificates(click_context: click.Context) -> None:
    """Issue certificates for connections that require them."""
    ctx = cast(Context, click_context.obj)
    agent_config_manager = AgentConfigManager.load(ctx.cwd)
    issue_certificates_(ctx.cwd, agent_config_manager)


def issue_certificates_(
    project_directory: str,
    agent_config_manager: AgentConfigManager,
    path_prefix: Optional[str] = None,
) -> None:
    """
    Issue certificates for connections that require them.

    :param project_directory: the directory of the project.
    :param agent_config_manager: the agent configuration manager.
    :param path_prefix: the path prefix for "save_path". Defaults to project directory.
    :return: None
    """
    path_prefix = path_prefix or project_directory
    for connection_id in agent_config_manager.agent_config.connections:
        cert_requests = _get_cert_requests(
            project_directory, agent_config_manager, connection_id
        )
        _process_connection(
            path_prefix, agent_config_manager, cert_requests, connection_id,
        )

    click.echo("All certificates have been issued.")


def _get_cert_requests(
    project_directory: str, manager: AgentConfigManager, connection_id: PublicId
) -> List[CertRequest]:
    """
    Get certificate requests, taking the overrides into account.

    :param project_directory: aea project directory.
    :param manager: AgentConfigManager
    :param connection_id: the connection id.

    :return: the list of cert requests.
    """
    path = get_dotted_package_path_unified(
        project_directory, manager.agent_config, CONNECTION, connection_id
    )
    path_to_cert_requests = f"{path}.cert_requests"

    try:
        cert_requests = manager.get_variable(path_to_cert_requests)
    except VariableDoesNotExist:
        return []

    cert_requests = cast(List[Dict], cert_requests)
    return [
        CertRequest.from_json(cert_request_json) for cert_request_json in cert_requests
    ]


def _process_certificate(
    path_prefix: str,
    agent_config: AgentConfig,
    cert_request: CertRequest,
    connection_id: PublicId,
) -> None:
    """Process a single certificate request."""
    ledger_id = cert_request.ledger_id
    if cert_request.key_identifier is not None:
        key_identifier = cert_request.key_identifier
        connection_private_key_path = agent_config.connection_private_key_paths.read(
            key_identifier
        )
        if connection_private_key_path is None:
            raise ClickException(
                f"Cannot find connection private key with id '{key_identifier}'. Connection '{connection_id}' requires this. Please use `aea generate-key {key_identifier} connection_{key_identifier}_private_key.txt` and `aea add-key {key_identifier} connection_{key_identifier}_private_key.txt --connection` to add a connection private key with id '{key_identifier}'."
            )
        new_connection_private_key_path = prepend_if_not_absolute(
            connection_private_key_path, path_prefix
        )
        connection_crypto = crypto_registry.make(
            key_identifier, private_key_path=new_connection_private_key_path
        )
        public_key = connection_crypto.public_key
    else:
        public_key = cast(str, cert_request.public_key)
        enforce(
            public_key is not None,
            "Internal error - one of key_identifier or public_key must be not None.",
        )
    crypto_private_key_path = agent_config.private_key_paths.read(ledger_id)
    if crypto_private_key_path is None:
        raise ClickException(
            f"Cannot find private key with id '{ledger_id}'. Please use `aea generate-key {key_identifier}` and `aea add-key {key_identifier}` to add a private key with id '{key_identifier}'."
        )
    message = cert_request.get_message(public_key)
    output_path = cert_request.get_absolute_save_path(path_prefix)
    absolute_crypto_private_key_path = prepend_if_not_absolute(
        crypto_private_key_path, path_prefix
    )
    cert = make_certificate(
        ledger_id, str(absolute_crypto_private_key_path), message, str(output_path),
    )
    click.echo(f"Generated signature: '{cert}'")
    click.echo(
        f"Dumped certificate '{cert_request.identifier}' in '{output_path}' for connection {connection_id}."
    )


def _process_connection(
    path_prefix: str,
    agent_config_manager: AgentConfigManager,
    cert_requests: List[CertRequest],
    connection_id: PublicId,
) -> None:

    if len(cert_requests) == 0:
        logger.debug("No certificates to process.")
        return

    logger.debug(f"Processing connection '{connection_id}'...")
    for cert_request in cert_requests:
        click.echo(
            f"Issuing certificate '{cert_request.identifier}' for connection {connection_id}..."
        )
        _process_certificate(
            path_prefix, agent_config_manager.agent_config, cert_request, connection_id,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the prometheus connection module."""
import asyncio
from typing import cast
from unittest.mock import MagicMock, Mock

import pytest

from aea.common import Address
from aea.configurations.base import ConnectionConfig, PublicId
from aea.exceptions import AEAEnforceError
from aea.identity.base import Identity
from aea.mail.base import Envelope, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue

from packages.fetchai.connections.prometheus.connection import (
    ConnectionStates,
    PrometheusConnection,
)
from packages.fetchai.protocols.prometheus.dialogues import PrometheusDialogue
from packages.fetchai.protocols.prometheus.dialogues import (
    PrometheusDialogues as BasePrometheusDialogues,
)
from packages.fetchai.protocols.prometheus.message import PrometheusMessage


class PrometheusDialogues(BasePrometheusDialogues):
    """The dialogues class keeps track of all prometheus dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return PrometheusDialogue.Role.AGENT

        BasePrometheusDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
        )


class TestPrometheusConnection:
    """Test the packages/connection/prometheus/connection.py."""

    def setup(self):
        """Initialise the class."""
        self.metrics = {}
        configuration = ConnectionConfig(
            connection_id=PrometheusConnection.connection_id, port=9090,
        )
        self.some_skill = "some/skill:0.1.0"
        self.agent_address = "my_address"
        self.protocol_specification_id = PublicId.from_str("fetchai/prometheus:0.4.0")
        identity = Identity("name", address=self.agent_address)
        self.prometheus_con = PrometheusConnection(
            identity=identity, configuration=configuration, data_dir=MagicMock()
        )
        self.loop = asyncio.get_event_loop()
        self.prometheus_address = str(PrometheusConnection.connection_id)
        self.dialogues = PrometheusDialogues(self.some_skill)

    async def send_add_metric(self, title: str, metric_type: str) -> None:
        """Send an add_metric message."""
        msg, sending_dialogue = self.dialogues.create(
            counterparty=self.prometheus_address,
            performative=PrometheusMessage.Performative.ADD_METRIC,
            title=title,
            type=metric_type,
            description="a gauge",
            labels={},
        )
        assert sending_dialogue is not None

        envelope = Envelope(to=msg.to, sender=msg.sender, message=msg,)
        await self.prometheus_con.send(envelope)

    async def send_update_metric(self, title: str, update_func: str) -> None:
        """Send an update_metric message."""
        msg, sending_dialogue = self.dialogues.create(
            counterparty=self.prometheus_address,
            performative=PrometheusMessage.Performative.UPDATE_METRIC,
            title=title,
            callable=update_func,
            value=1.0,
            labels={},
        )
        assert sending_dialogue is not None
        assert sending_dialogue.last_message is not None

        envelope = Envelope(to=msg.to, sender=msg.sender, message=msg,)
        await self.prometheus_con.send(envelope)

    def teardown(self):
        """Clean up after tests."""
        self.loop.run_until_complete(self.prometheus_con.disconnect())

    @pytest.mark.asyncio
    async def test_connection(self):
        """Test connect."""
        assert (
            self.prometheus_con.state == ConnectionStates.disconnected
        ), "should not be connected yet"
        await self.prometheus_con.connect()
        assert (
            self.prometheus_con.state == ConnectionStates.connected
        ), "should be connected"

        # test add metric (correct)
        await self.send_add_metric("some_metric", "Gauge")
        envelope = await self.prometheus_con.receive()
        msg = cast(PrometheusMessage, envelope.message)
        assert msg.performative == PrometheusMessage.Performative.RESPONSE
        assert msg.code == 200
        assert msg.message == "New Gauge successfully added: some_metric."

        # test add metric (already exists)
        await self.send_add_metric("some_metric", "Gauge")
        envelope = await self.prometheus_con.receive()
        msg = cast(PrometheusMessage, envelope.message)
        assert msg.performative == PrometheusMessage.Performative.RESPONSE
        assert msg.code == 409
        assert msg.message == "Metric already exists."

        # test add metric (wrong type)
        await self.send_add_metric("cool_metric", "CoolBar")
        envelope = await self.prometheus_con.receive()
        msg = cast(PrometheusMessage, envelope.message)
        assert msg.performative == PrometheusMessage.Performative.RESPONSE
        assert msg.code == 404
        assert msg.message == "CoolBar is not a recognized prometheus metric."

        # test update metric (inc: correct)
        await self.send_update_metric("some_metric", "inc")
        envelope = await self.prometheus_con.receive()
        msg = cast(PrometheusMessage, envelope.message)
        assert msg.performative == PrometheusMessage.Performative.RESPONSE
        assert msg.code == 200
        assert msg.message == "Metric some_metric successfully updated."

        # test update metric (set: correct)
        await self.send_update_metric("some_metric", "set")
        envelope = await self.prometheus_con.receive()
        msg = cast(PrometheusMessage, envelope.message)
        assert msg.performative == PrometheusMessage.Performative.RESPONSE
        assert msg.code == 200
        assert msg.message == "Metric some_metric successfully updated."

        # test update metric (doesn't exist)
        await self.send_update_metric("cool_metric", "inc")
        envelope = await self.prometheus_con.receive()
        msg = cast(PrometheusMessage, envelope.message)
        assert msg.performative == PrometheusMessage.Performative.RESPONSE
        assert msg.code == 404
        assert msg.message == "Metric cool_metric not found."

        # test update metric (bad update function: not found in attr)
        await self.send_update_metric("some_metric", "go")
        envelope = await self.prometheus_con.receive()
        msg = cast(PrometheusMessage, envelope.message)
        assert msg.performative == PrometheusMessage.Performative.RESPONSE
        assert msg.code == 400
        assert msg.message == "Update function go not found for metric some_metric."

        # test update metric (bad update function: found in getattr, not a method)
        await self.send_update_metric("some_metric", "name")
        envelope = await self.prometheus_con.receive()
        msg = cast(PrometheusMessage, envelope.message)
        assert msg.performative == PrometheusMessage.Performative.RESPONSE
        assert msg.code == 400
        assert (
            msg.message
            == "Failed to update metric some_metric: name is not a valid update function."
        )

        # Test that invalid message is rejected.
        with pytest.raises(AEAEnforceError):
            envelope = Envelope(
                to="some_address", sender="me", message=Mock(spec=Message),
            )
            await self.prometheus_con.channel.send(envelope)

        # Test that envelope without dialogue produces warning.
        msg = PrometheusMessage(
            PrometheusMessage.Performative.RESPONSE, code=0, message=""
        )
        envelope = Envelope(
            to=self.prometheus_address, sender=self.some_skill, message=msg,
        )
        await self.prometheus_con.channel.send(envelope)

        # Test that envelope with invalid protocol_specification_id raises error.
        with pytest.raises(ValueError):
            msg, _ = self.dialogues.create(
                counterparty=self.prometheus_address,
                performative=PrometheusMessage.Performative.UPDATE_METRIC,
                title="",
                callable="",
                value=1.0,
                labels={},
            )
            envelope = Envelope(
                to=self.prometheus_address, sender=self.some_skill, message=msg,
            )
            envelope._protocol_specification_id = "bad_id"
            await self.prometheus_con.channel.send(envelope)

    @pytest.mark.asyncio
    async def test_disconnect(self):
        """Test disconnect."""
        await self.prometheus_con.disconnect()
        assert (
            self.prometheus_con.state == ConnectionStates.disconnected
        ), "should be disconnected"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the content of query-language.md file."""
from pathlib import Path
from unittest import mock
from unittest.mock import MagicMock

from tests.conftest import ROOT_DIR
from tests.test_docs.helper import BasePythonMarkdownDocs


@mock.patch("aea.helpers.search.models.Query.check_validity")
class TestQueryLanguage(BasePythonMarkdownDocs):
    """Test the data models code snippets."""

    DOC_PATH = Path(ROOT_DIR, "docs", "query-language.md")

    @classmethod
    def setup_class(cls):
        """Set up the test."""
        super().setup_class()
        cls.locals["book_model"] = MagicMock()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour classes of the simple oracle skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.contracts.oracle.contract import PUBLIC_ID as CONTRACT_PUBLIC_ID
from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage
from packages.fetchai.protocols.prometheus.message import PrometheusMessage
from packages.fetchai.skills.simple_oracle.behaviours import SimpleOracleBehaviour
from packages.fetchai.skills.simple_oracle.dialogues import PrometheusDialogues
from packages.fetchai.skills.simple_oracle.strategy import Strategy

from tests.conftest import ROOT_DIR


LEDGER_ID = "ethereum"
AGENT_ADDRESS = "some_eth_address"
DEFAULT_ADDRESS = "0x0000000000000000000000000000000000000000"


class TestSkillBehaviour(BaseSkillTestCase):
    """Test behaviours of simple oracle."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "simple_oracle")

    @classmethod
    def setup(cls, **kwargs):
        """Setup the test class."""
        super().setup()
        cls.simple_oracle_behaviour = cast(
            SimpleOracleBehaviour,
            cls._skill.skill_context.behaviours.simple_oracle_behaviour,
        )

    def test_setup(self):
        """Test the setup method of the simple_oracle behaviour."""
        self.simple_oracle_behaviour.setup()
        self.assert_quantity_in_outbox(3)

        msg = cast(ContractApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            contract_id=str(CONTRACT_PUBLIC_ID),
            callable="get_deploy_transaction",
        )
        assert has_attributes, error_str

        msg = cast(PrometheusMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.ADD_METRIC,
            type="Gauge",
            title="oracle_account_balance_ETH",
            description="Balance of oracle contract (ETH)",
        )
        assert has_attributes, error_str

        msg = cast(PrometheusMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.ADD_METRIC,
            type="Gauge",
            title="num_oracle_updates",
            description="Number of updates published to oracle contract",
        )
        assert has_attributes, error_str

    def test_setup_with_contract_set(self):
        """Test the setup method of the simple_oracle behaviour for existing contract."""
        prometheus_dialogues = cast(
            PrometheusDialogues,
            self.simple_oracle_behaviour.context.prometheus_dialogues,
        )
        prometheus_dialogues.enabled = False
        strategy = cast(Strategy, self.simple_oracle_behaviour.context.strategy)
        strategy.contract_address = DEFAULT_ADDRESS
        strategy.is_contract_deployed = True

        with patch.object(
            self.simple_oracle_behaviour.context.logger, "log"
        ) as mock_logger:
            self.simple_oracle_behaviour.setup()
        mock_logger.assert_any_call(
            logging.INFO, "Fetch oracle contract address already added",
        )
        self.assert_quantity_in_outbox(0)

    def test_setup_with_contract_set_and_oracle_role_granted(self):
        """Test the setup method of the simple_oracle behaviour for existing contract and oracle role."""
        prometheus_dialogues = cast(
            PrometheusDialogues,
            self.simple_oracle_behaviour.context.prometheus_dialogues,
        )
        prometheus_dialogues.enabled = False
        strategy = cast(Strategy, self.simple_oracle_behaviour.context.strategy)
        strategy.contract_address = DEFAULT_ADDRESS
        strategy.is_contract_deployed = True
        strategy.is_oracle_role_granted = True

        assert strategy.erc20_address == DEFAULT_ADDRESS

        with patch.object(
            self.simple_oracle_behaviour.context.logger, "log"
        ) as mock_logger:
            self.simple_oracle_behaviour.setup()
        mock_logger.assert_any_call(
            logging.INFO, "Oracle role already granted",
        )

        self.assert_quantity_in_outbox(0)

    def test_act_pre_deploy(self):
        """Test the act method of the simple_oracle behaviour before contract is deployed."""
        self.simple_oracle_behaviour.context.agent_addresses[
            LEDGER_ID
        ] = "AGENT_ADDRESS"
        self.simple_oracle_behaviour.act()
        self.assert_quantity_in_outbox(1)

        msg = cast(LedgerApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id=LEDGER_ID,
            address=cast(
                str,
                self.simple_oracle_behaviour.context.agent_addresses.get(LEDGER_ID),
            ),
        )
        assert has_attributes, error_str

    def test_act_grant_role(self):
        """Test the act method of the simple_oracle behaviour before role is granted."""
        strategy = cast(Strategy, self.simple_oracle_behaviour.context.strategy)
        strategy.contract_address = DEFAULT_ADDRESS
        strategy.is_contract_deployed = True
        self.simple_oracle_behaviour.context.agent_addresses[
            LEDGER_ID
        ] = "AGENT_ADDRESS"
        self.simple_oracle_behaviour.act()
        self.assert_quantity_in_outbox(2)

        msg = cast(LedgerApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id=LEDGER_ID,
            address=cast(
                str,
                self.simple_oracle_behaviour.context.agent_addresses.get(LEDGER_ID),
            ),
        )
        assert has_attributes, error_str

        msg = cast(ContractApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            contract_id=str(CONTRACT_PUBLIC_ID),
            contract_address=strategy.contract_address,
            callable="get_grant_role_transaction",
        )
        assert has_attributes, error_str

    def test_act_update(self):
        """Test the act method of the simple_oracle behaviour for normal updating."""
        strategy = cast(Strategy, self.simple_oracle_behaviour.context.strategy)
        strategy.contract_address = DEFAULT_ADDRESS
        strategy.is_contract_deployed = True
        strategy.is_oracle_role_granted = True
        self.simple_oracle_behaviour.context.shared_state["oracle_data"] = {
            "some_key": "some_value"
        }
        self.simple_oracle_behaviour.context.agent_addresses[
            LEDGER_ID
        ] = "AGENT_ADDRESS"
        self.simple_oracle_behaviour.act()
        self.assert_quantity_in_outbox(2)

        msg = cast(LedgerApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id=LEDGER_ID,
            address=cast(
                str,
                self.simple_oracle_behaviour.context.agent_addresses.get(LEDGER_ID),
            ),
        )
        assert has_attributes, error_str

        msg = cast(ContractApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            contract_id=str(CONTRACT_PUBLIC_ID),
            contract_address=strategy.contract_address,
            callable="get_update_transaction",
        )
        assert has_attributes, error_str

    def test_act_no_oracle_value(self):
        """Test the act method of the simple_oracle behaviour when no oracle value is present."""
        strategy = cast(Strategy, self.simple_oracle_behaviour.context.strategy)
        strategy.contract_address = DEFAULT_ADDRESS
        strategy.is_contract_deployed = True
        strategy.is_oracle_role_granted = True
        self.simple_oracle_behaviour.context.agent_addresses[
            LEDGER_ID
        ] = "AGENT_ADDRESS"
        self.simple_oracle_behaviour.act()
        self.assert_quantity_in_outbox(1)

        msg = cast(LedgerApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id=LEDGER_ID,
            address=cast(
                str,
                self.simple_oracle_behaviour.context.agent_addresses.get(LEDGER_ID),
            ),
        )
        assert has_attributes, error_str

    def test__request_contract_deploy_transaction(self):
        """Test that the _request_contract_deploy_transaction function sends the right message to the contract_api."""
        self.simple_oracle_behaviour._request_contract_deploy_transaction()
        self.assert_quantity_in_outbox(1)

        msg = cast(ContractApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            contract_id=str(CONTRACT_PUBLIC_ID),
            callable="get_deploy_transaction",
        )
        assert has_attributes, error_str

    def test__request_grant_role_transaction(self):
        """Test that the _request_grant_role_transaction function sends the right message to the contract_api."""
        strategy = cast(Strategy, self.simple_oracle_behaviour.context.strategy)
        strategy.contract_address = DEFAULT_ADDRESS

        self.simple_oracle_behaviour._request_grant_role_transaction()
        self.assert_quantity_in_outbox(1)

        msg = cast(ContractApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            contract_id=str(CONTRACT_PUBLIC_ID),
            contract_address=strategy.contract_address,
            callable="get_grant_role_transaction",
        )
        assert has_attributes, error_str

    def test__request_update_transaction(self):
        """Test that the _request_update_transaction function sends the right message to the contract_api."""
        strategy = cast(Strategy, self.simple_oracle_behaviour.context.strategy)
        strategy.contract_address = DEFAULT_ADDRESS

        update_args = {"some": "args"}
        self.simple_oracle_behaviour._request_update_transaction(update_args)
        self.assert_quantity_in_outbox(1)

        msg = cast(ContractApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            contract_id=str(CONTRACT_PUBLIC_ID),
            contract_address=strategy.contract_address,
            callable="get_update_transaction",
        )
        assert has_attributes, error_str

    def test__get_balance(self):
        """Test that the _get_balance function sends the right message to the ledger_api."""
        self.simple_oracle_behaviour.context.agent_addresses[
            LEDGER_ID
        ] = "AGENT_ADDRESS"

        self.simple_oracle_behaviour._get_balance()
        self.assert_quantity_in_outbox(1)

        msg = cast(LedgerApiMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id=LEDGER_ID,
            address=cast(
                str,
                self.simple_oracle_behaviour.context.agent_addresses.get(LEDGER_ID),
            ),
        )
        assert has_attributes, error_str

    def test_add_prometheus_metric(self):
        """Test the send_http_request_message method of the simple_oracle behaviour."""
        self.simple_oracle_behaviour.add_prometheus_metric(
            "some_metric", "Gauge", "some_description", {"label_key": "label_value"}
        )
        self.assert_quantity_in_outbox(1)

        msg = cast(PrometheusMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.ADD_METRIC,
            type="Gauge",
            title="some_metric",
            description="some_description",
            labels={"label_key": "label_value"},
        )
        assert has_attributes, error_str

    def test_update_prometheus_metric(self):
        """Test the test_update_prometheus_metric method of the simple_oracle behaviour."""
        self.simple_oracle_behaviour.update_prometheus_metric(
            "some_metric", "set", 0.0, {"label_key": "label_value"}
        )
        self.assert_quantity_in_outbox(1)

        msg = cast(PrometheusMessage, self.get_message_from_outbox())
        has_attributes, error_str = self.message_has_attributes(
            actual_message=msg,
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.UPDATE_METRIC,
            callable="set",
            title="some_metric",
            value=0.0,
            labels={"label_key": "label_value"},
        )
        assert has_attributes, error_str

    def test_teardown(self):
        """Test that the teardown method of the simple_oracle behaviour leaves no messages in the outbox."""
        assert self.simple_oracle_behaviour.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the simple_data_request skill."""

import logging
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.protocols.dialogue.base import DialogueMessage
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.connections.http_server.connection import (
    PUBLIC_ID as HTTP_SERVER_ID,
)
from packages.fetchai.protocols.http.message import HttpMessage
from packages.fetchai.protocols.prometheus.message import PrometheusMessage
from packages.fetchai.skills.coin_price.dialogues import (
    HttpDialogues,
    PrometheusDialogues,
)
from packages.fetchai.skills.coin_price.handlers import HttpHandler, PrometheusHandler
from packages.fetchai.skills.coin_price.models import CoinPriceModel

from tests.conftest import ROOT_DIR


class TestHttpHandler(BaseSkillTestCase):
    """Test http handler of coin_price skill."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "coin_price")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls, **kwargs):
        """Setup the test class."""
        super().setup()
        cls.http_handler = cast(HttpHandler, cls._skill.skill_context.handlers.http)
        cls.logger = cls._skill.skill_context.logger

        cls.coin_price_model = cast(
            CoinPriceModel, cls._skill.skill_context.coin_price_model
        )

        cls.http_dialogues = cast(
            HttpDialogues, cls._skill.skill_context.http_dialogues
        )

        cls.list_of_messages = (
            DialogueMessage(
                HttpMessage.Performative.REQUEST,
                {
                    "method": "get",
                    "url": "some_url",
                    "headers": "",
                    "version": "",
                    "body": b"",
                },
            ),
        )

    def test_setup(self):
        """Test the setup method of the http handler."""
        assert self.http_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_setup_with_http_server(self):
        """Test the setup method of the http handler."""
        self.coin_price_model.use_http_server = True
        assert self.http_handler.setup() is None

        assert self.http_handler._http_server_id == HTTP_SERVER_ID
        self.assert_quantity_in_outbox(0)

    def test_handle_response(self):
        """Test the _handle_response method of the http handler to a valid coin price response."""
        # setup
        http_dialogue = self.prepare_skill_dialogue(
            dialogues=self.http_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=http_dialogue,
            performative=HttpMessage.Performative.RESPONSE,
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=b'{"fetch-ai":{"usd":100.00}}',
        )

        # handle message
        self.http_handler.handle(incoming_message)

        # check that data was correctly entered into shared state
        oracle_data = {"value": 10000000, "decimals": self.coin_price_model.decimals}
        assert self.http_handler.context.shared_state["oracle_data"] == oracle_data

        # check that outbox contains update_prometheus metric message
        self.assert_quantity_in_outbox(1)

    def test_handle__handle_unidentified_dialogue(self):
        """Test handling an unidentified dialogoue"""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=HttpMessage.Performative.RESPONSE,
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=b"{}",
        )

        # operation
        with patch.object(self.http_handler.context.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid message={incoming_message}, unidentified dialogue.",
        )

        self.assert_quantity_in_outbox(0)

    def test_handle_response_invalid_body(self):
        """Test the _handle_response method of the http handler to an unexpected response."""
        # setup
        http_dialogue = self.prepare_skill_dialogue(
            dialogues=self.http_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=http_dialogue,
            performative=HttpMessage.Performative.RESPONSE,
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=b"{}",
        )

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        assert "oracle_data" not in self.http_handler.context.shared_state

        # after
        mock_logger.assert_any_call(
            logging.INFO, "failed to get price: unexpected result",
        )

    def test_handle_response_no_price(self):
        """Test the _handle_response method of the http handler to a response with no price."""
        # setup
        http_dialogue = self.prepare_skill_dialogue(
            dialogues=self.http_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=http_dialogue,
            performative=HttpMessage.Performative.RESPONSE,
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=b'{"fetch-ai":{}}',
        )

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        assert "oracle_data" not in self.http_handler.context.shared_state

        # after
        mock_logger.assert_any_call(
            logging.INFO, "failed to get price: no price listed",
        )

    def test_handle_response_bad_response_code(self):
        """Test the _handle_response method of the http handler to a response with a code that is not 200."""
        # setup
        http_dialogue = self.prepare_skill_dialogue(
            dialogues=self.http_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=http_dialogue,
            performative=HttpMessage.Performative.RESPONSE,
            version="",
            status_code=999,
            status_text="",
            headers="",
            body=b'{"fetch-ai":{}}',
        )

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        assert "oracle_data" not in self.http_handler.context.shared_state

        # after
        mock_logger.assert_any_call(
            logging.INFO, "got unexpected http message: code = 999",
        )

    def test_handle_request_get(self):
        """Test the _handle_request method of the http handler with 'get' method."""

        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            performative=HttpMessage.Performative.REQUEST,
            method="get",
            url="some_url",
            headers="",
            version="",
            body=b"",
        )

        self.http_handler._http_server_id = "some_id"

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            "received http request with method=get, url=some_url and body=b''",
        )

        # check that outbox contains the http response prometheus metric update messages
        self.assert_quantity_in_outbox(2)

    def test_handle_request_post(self):
        """Test the _handle_request method of the http handler with 'post' method."""

        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            performative=HttpMessage.Performative.REQUEST,
            method="post",
            url="some_url",
            headers="",
            version="",
            body=b"",
        )

        self.http_handler._http_server_id = "some_id"

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        mock_logger.assert_any_call(
            logging.INFO, "method 'post' is not supported.",
        )
        # check that outbox is empty
        self.assert_quantity_in_outbox(0)

    def test_handle_request_no_http_server(self):
        """Test the _handle_request method of the http handler when http server is disabled."""

        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            performative=HttpMessage.Performative.REQUEST,
            method="get",
            url="some_url",
            headers="",
            version="",
            body=b"",
        )

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.http_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            "received http request with method=get, url=some_url and body=b''",
        )

        mock_logger.assert_any_call(
            logging.INFO, "http server is not enabled.",
        )
        # check that outbox is empty
        self.assert_quantity_in_outbox(0)

    def test_teardown(self):
        """Test the teardown method of the http handler."""
        assert self.http_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestPrometheusHandler(BaseSkillTestCase):
    """Test prometheus handler of coin_price skill."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "coin_price")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls, **kwargs):
        """Setup the test class."""
        super().setup()
        cls.prometheus_handler = cast(
            PrometheusHandler, cls._skill.skill_context.handlers.prometheus
        )
        cls.logger = cls._skill.skill_context.logger

        cls.coin_price_model = cast(
            CoinPriceModel, cls._skill.skill_context.coin_price_model
        )

        cls.prometheus_dialogues = cast(
            PrometheusDialogues, cls._skill.skill_context.prometheus_dialogues
        )

        cls.list_of_messages = (
            DialogueMessage(
                PrometheusMessage.Performative.ADD_METRIC,
                {
                    "type": "Gauge",
                    "title": "some_title",
                    "description": "some_description",
                    "labels": {},
                },
            ),
        )

    def test_setup(self):
        """Test the setup method of the prometheus handler."""
        assert self.prometheus_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_response(self):
        """Test the _handle_response method of the prometheus handler to a valid coin price response."""
        # setup
        prometheus_dialogue = self.prepare_skill_dialogue(
            dialogues=self.prometheus_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=prometheus_dialogue,
            performative=PrometheusMessage.Performative.RESPONSE,
            code=200,
            message="some_message",
        )

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.prometheus_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.DEBUG, "Prometheus response (200): some_message"
        )

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_response method of the prometheus handler to an unidentified dialogue."""
        # setup
        incoming_message = self.build_incoming_message(
            message_type=PrometheusMessage,
            performative=PrometheusMessage.Performative.RESPONSE,
            code=200,
            message="some_message",
        )

        # handle message with logging
        with patch.object(self.logger, "log") as mock_logger:
            self.prometheus_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid message={incoming_message}, unidentified dialogue.",
        )

    def test_teardown(self):
        """Test the teardown method of the prometheus handler."""
        assert self.prometheus_handler.teardown() is None
        self.assert_quantity_in_outbox(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the integration test for the fetch beacon skill."""

import pytest

from aea.test_tools.test_cases import AEATestCaseEmpty


@pytest.mark.integration
class TestFetchBeaconSkill(AEATestCaseEmpty):
    """Test that fetch beacon skill works."""

    def test_fetch_beacon(self):
        """Run the fetch beacon skill sequence."""
        self.add_item("connection", "fetchai/http_client:0.18.0")
        self.add_item("skill", "fetchai/fetch_beacon:0.6.0")
        self.set_config("agent.default_connection", "fetchai/http_client:0.18.0")

        self.run_install()

        process = self.run_agent()
        is_running = self.is_running(process)
        assert is_running, "AEA not running within timeout!"

        check_strings = (
            "setting up FetchBeaconBehaviour",
            "Fetching random beacon from https://rpc-beaconworld.fetch.ai/block...",
            "Beacon info:",
            "entropy not present",
        )
        missing_strings = self.missing_from_output(process, check_strings)
        assert len(missing_strings) in [
            0,
            1,
        ], "Strings {} didn't appear in agent output.".format(missing_strings)

        self.terminate_agents()
        assert (
            self.is_successfully_terminated()
        ), "Http echo agent wasn't successfully terminated."



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains a test for aea.test_tools.test_cases."""

from pathlib import Path
from typing import cast

import pytest

from aea.decision_maker.gop import GoalPursuitReadiness, OwnershipState, Preferences
from aea.exceptions import AEAEnforceError
from aea.mail.base import Address
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue, DialogueLabel, DialogueMessage
from aea.skills.base import Skill
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.fipa.dialogues import FipaDialogue
from packages.fetchai.protocols.fipa.dialogues import FipaDialogues as BaseFipaDialogues
from packages.fetchai.protocols.fipa.message import FipaMessage

from tests.conftest import ROOT_DIR


class TestSkillTestCase(BaseSkillTestCase):
    """Test case for BaseSkillTestCase."""

    path_to_skill = Path(ROOT_DIR, "tests", "data", "dummy_skill")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        cls.behaviour_arg_1 = 2
        cls.behaviour_arg_2 = "3"

        config_overrides = {
            "behaviours": {
                "dummy": {
                    "args": {
                        "behaviour_arg_1": cls.behaviour_arg_1,
                        "behaviour_arg_2": cls.behaviour_arg_2,
                    }
                }
            },
        }
        cls.shared_state_key = "some_shared_state_key"
        cls.shared_state_value = "some_shared_state_value"
        cls.shared_state = {cls.shared_state_key: cls.shared_state_value}

        tac_dm_context_kwargs = {
            "goal_pursuit_readiness": GoalPursuitReadiness(),
            "ownership_state": OwnershipState(),
            "preferences": Preferences(),
        }

        super().setup(
            config_overrides=config_overrides,
            shared_state=cls.shared_state,
            dm_context_kwargs=tac_dm_context_kwargs,
        )

    def test_setup(self):
        """Test the setup() class method."""
        assert self.skill.skill_context.agent_address == "test_agent_address"
        assert self.skill.skill_context.agent_name == "test_agent_name"
        assert (
            self.skill.skill_context.search_service_address
            == "dummy_author/dummy_search_skill:0.1.0"
        )
        assert (
            self.skill.skill_context.decision_maker_address
            == "dummy_decision_maker_address"
        )
        assert "dummy" in self.skill.behaviours.keys()
        assert "dummy" in self.skill.handlers.keys()
        assert "dummy_internal" in self.skill.handlers.keys()
        assert "dummy" in self.skill.models.keys()

        assert (
            self.skill.skill_context._agent_context.shared_state[self.shared_state_key]
            == self.shared_state_value
        )

        assert (
            self.skill.skill_context.behaviours.dummy.kwargs["behaviour_arg_1"]
            == self.behaviour_arg_1
        )
        assert (
            self.skill.skill_context.behaviours.dummy.kwargs["behaviour_arg_2"]
            == self.behaviour_arg_2
        )

        assert (
            type(
                self.skill.skill_context.decision_maker_handler_context.goal_pursuit_readiness
            )
            == GoalPursuitReadiness
        )
        assert (
            type(
                self.skill.skill_context.decision_maker_handler_context.ownership_state
            )
            == OwnershipState
        )
        assert (
            type(self.skill.skill_context.decision_maker_handler_context.preferences)
            == Preferences
        )

    def test_properties(self):
        """Test the properties."""
        assert isinstance(self.skill, Skill)
        assert self.skill.behaviours.get("dummy") is not None

    def test_get_quantity_in_outbox(self):
        """Test the get_quantity_in_outbox method."""
        assert self.get_quantity_in_outbox() == 0

        dummy_message = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy"
        )
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"
        self.skill.skill_context.outbox.put_message(dummy_message)

        assert self.get_quantity_in_outbox() == 1

    def test_get_message_from_outbox(self):
        """Test the get_message_from_outbox method."""
        assert self.get_message_from_outbox() is None

        dummy_message_1 = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy_1"
        )
        dummy_message_1.to = "some_to_1"
        dummy_message_1.sender = "some_sender_1"
        self.skill.skill_context.outbox.put_message(dummy_message_1)

        dummy_message_2 = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy_2"
        )
        dummy_message_2.to = "some_to_2"
        dummy_message_2.sender = "some_sender_2"
        self.skill.skill_context.outbox.put_message(dummy_message_2)

        assert self.get_message_from_outbox() == dummy_message_1
        assert self.get_message_from_outbox() == dummy_message_2

    def test_drop_messages_from_outbox(self):
        """Test the drop_messages_from_outbox method."""
        assert self.get_quantity_in_outbox() == 0
        self.drop_messages_from_outbox(5)
        assert self.get_quantity_in_outbox() == 0

        dummy_message_1 = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy_2"
        )
        dummy_message_1.to = "some_to_1"
        dummy_message_1.sender = "some_sender_1"
        self.skill.skill_context.outbox.put_message(dummy_message_1)

        dummy_message_2 = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy_2"
        )
        dummy_message_2.to = "some_to_2"
        dummy_message_2.sender = "some_sender_2"
        self.skill.skill_context.outbox.put_message(dummy_message_2)

        dummy_message_3 = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy_2"
        )
        dummy_message_3.to = "some_to_3"
        dummy_message_3.sender = "some_sender_3"
        self.skill.skill_context.outbox.put_message(dummy_message_3)

        assert self.get_quantity_in_outbox() == 3

        self.drop_messages_from_outbox(2)

        assert self.get_quantity_in_outbox() == 1
        assert self.get_message_from_outbox() == dummy_message_3

    def test_get_quantity_in_decision_maker_inbox(self):
        """Test the get_quantity_in_decision_maker_inbox method."""
        assert self.get_quantity_in_decision_maker_inbox() == 0

        dummy_message = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy"
        )
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message)

        assert self.get_quantity_in_decision_maker_inbox() == 1

    def test_get_message_from_decision_maker_inbox(self):
        """Test the get_message_from_decision_maker_inbox method."""
        assert self.get_message_from_decision_maker_inbox() is None

        dummy_message_1 = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy_1"
        )
        dummy_message_1.to = "some_to_1"
        dummy_message_1.sender = "some_sender_1"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message_1)

        dummy_message_2 = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy_2"
        )
        dummy_message_2.to = "some_to_2"
        dummy_message_2.sender = "some_sender_2"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message_2)

        assert self.get_message_from_decision_maker_inbox() == dummy_message_1
        assert self.get_message_from_decision_maker_inbox() == dummy_message_2

    def test_drop_messages_from_decision_maker_inbox(self):
        """Test the drop_messages_from_decision_maker_inbox method."""
        assert self.get_quantity_in_decision_maker_inbox() == 0
        self.drop_messages_from_decision_maker_inbox(5)
        assert self.get_quantity_in_decision_maker_inbox() == 0

        dummy_message_1 = Message()
        dummy_message_1.to = "some_to_1"
        dummy_message_1.sender = "some_sender_1"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message_1)

        dummy_message_2 = Message()
        dummy_message_2.to = "some_to_2"
        dummy_message_2.sender = "some_sender_2"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message_2)

        dummy_message_3 = Message()
        dummy_message_3.to = "some_to_3"
        dummy_message_3.sender = "some_sender_3"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message_3)

        assert self.get_quantity_in_decision_maker_inbox() == 3

        self.drop_messages_from_decision_maker_inbox(2)

        assert self.get_quantity_in_decision_maker_inbox() == 1
        assert self.get_message_from_decision_maker_inbox() == dummy_message_3

    def test_assert_quantity_in_outbox(self):
        """Test the assert_quantity_in_outbox method."""
        with pytest.raises(
            AssertionError,
            match=f"Invalid number of messages in outbox. Expected {1}. Found {0}.",
        ):
            self.assert_quantity_in_outbox(1)

        dummy_message = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy"
        )
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"
        self.skill.skill_context.outbox.put_message(dummy_message)

        self.assert_quantity_in_outbox(1)

    def test_assert_quantity_in_decision_making_queue(self):
        """Test the assert_quantity_in_decision_making_queue method."""
        with pytest.raises(
            AssertionError,
            match=f"Invalid number of messages in decision maker queue. Expected {1}. Found {0}.",
        ):
            self.assert_quantity_in_decision_making_queue(1)

        dummy_message = DefaultMessage(
            performative=DefaultMessage.Performative.BYTES, content="dummy_1"
        )
        dummy_message.to = "some_to_1"
        dummy_message.sender = "some_sender_1"
        self.skill.skill_context.decision_maker_message_queue.put(dummy_message)

        self.assert_quantity_in_decision_making_queue(1)

    def test_positive_message_has_attributes_valid_type(self):
        """Test the message_has_attributes method where the message is of the specified type."""
        dummy_message = FipaMessage(
            dialogue_reference=("0", "0"),
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
        )
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"

        valid_has_attribute, valid_has_attribute_msg = self.message_has_attributes(
            actual_message=dummy_message,
            message_type=FipaMessage,
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
            to="some_to",
            sender="some_sender",
        )
        assert valid_has_attribute
        assert (
            valid_has_attribute_msg
            == "The message has the provided expected attributes."
        )

    def test_negative_message_has_attributes_invalid_message_id(self):
        """Negative test for message_has_attributes method where the message id does NOT match."""
        dummy_message = FipaMessage(
            dialogue_reference=("0", "0"),
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
        )
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"

        invalid_has_attribute, invalid_has_attribute_msg = self.message_has_attributes(
            actual_message=dummy_message,
            message_type=FipaMessage,
            message_id=2,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
            to="some_to",
            sender="some_sender",
        )
        assert not invalid_has_attribute
        assert (
            invalid_has_attribute_msg
            == "The 'message_id' fields do not match. Actual 'message_id': 1. Expected 'message_id': 2"
        )

    def test_negative_message_has_attributes_invalid_type(self):
        """Test the message_has_attributes method where the message is NOT of the specified type."""
        dummy_message = FipaMessage(
            dialogue_reference=("0", "0"),
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
        )
        dummy_message.to = "some_to"
        dummy_message.sender = "some_sender"

        valid_has_attribute, valid_has_attribute_msg = self.message_has_attributes(
            actual_message=dummy_message,
            message_type=DefaultMessage,
            message_id=1,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
            to="some_to",
            sender="some_sender",
        )
        assert not valid_has_attribute
        assert (
            valid_has_attribute_msg
            == f"The message types do not match. Actual type: {FipaMessage}. Expected type: {DefaultMessage}"
        )

        invalid_has_attribute, invalid_has_attribute_msg = self.message_has_attributes(
            actual_message=dummy_message,
            message_type=FipaMessage,
            message_id=2,
            performative=FipaMessage.Performative.CFP,
            target=0,
            query="some_query",
            to="some_to",
            sender="some_sender",
        )
        assert not invalid_has_attribute
        assert (
            invalid_has_attribute_msg
            == "The 'message_id' fields do not match. Actual 'message_id': 1. Expected 'message_id': 2"
        )

    def test_build_incoming_message(self):
        """Test the build_incoming_message method."""
        message_type = FipaMessage
        performative = FipaMessage.Performative.CFP
        dialogue_reference = ("1", "1")
        to = "some_to"
        query = "some_query"
        incoming_message = self.build_incoming_message(
            message_type=message_type,
            performative=performative,
            dialogue_reference=dialogue_reference,
            to=to,
            query=query,
        )

        assert type(incoming_message) == message_type
        incoming_message = cast(FipaMessage, incoming_message)
        assert incoming_message.dialogue_reference == dialogue_reference
        assert incoming_message.message_id == 1
        assert incoming_message.target == 0
        assert incoming_message.performative == performative
        assert incoming_message.query == query
        assert incoming_message.sender == "counterparty"
        assert incoming_message.to == to

    def test_positive_build_incoming_message_for_skill_dialogue(self):
        """Positive test for build_incoming_message_for_skill_dialogue method."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        base_msg, dialogue = fipa_dialogues.create(
            counterparty="some_counterparty",
            performative=FipaMessage.Performative.CFP,
            query="some_query",
        )

        performative = FipaMessage.Performative.PROPOSE
        proposal = "some_proposal"
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=dialogue, performative=performative, proposal=proposal,
        )

        assert type(incoming_message) == FipaMessage
        incoming_message = cast(FipaMessage, incoming_message)
        assert (
            incoming_message.dialogue_reference
            == dialogue.dialogue_label.dialogue_reference
        )
        assert incoming_message.message_id != base_msg.message_id
        assert incoming_message.target == base_msg.message_id
        assert incoming_message.performative == performative
        assert incoming_message.proposal == proposal
        assert incoming_message.sender == dialogue.dialogue_label.dialogue_opponent_addr
        assert incoming_message.to == dialogue.self_address

    def test_negative_build_incoming_message_for_skill_dialogue_dialogue_is_none(self):
        """Negative test for build_incoming_message_for_skill_dialogue method where the provided dialogue is None."""
        performative = FipaMessage.Performative.PROPOSE
        proposal = "some_proposal"

        with pytest.raises(AEAEnforceError, match="dialogue cannot be None."):
            self.build_incoming_message_for_skill_dialogue(
                dialogue=None, performative=performative, proposal=proposal,
            )

    def test_negative_build_incoming_message_for_skill_dialogue_dialogue_is_empty(self):
        """Negative test for build_incoming_message_for_skill_dialogue method where the provided dialogue is empty."""
        performative = FipaMessage.Performative.PROPOSE
        proposal = "some_proposal"

        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue = fipa_dialogues._create_self_initiated(
            dialogue_opponent_addr="some_counterparty",
            dialogue_reference=("0", ""),
            role=FipaDialogue.Role.BUYER,
        )

        with pytest.raises(AEAEnforceError, match="dialogue cannot be empty."):
            self.build_incoming_message_for_skill_dialogue(
                dialogue=dialogue, performative=performative, proposal=proposal,
            )

    def test_provide_unspecified_fields(self):
        """Test the _provide_unspecified_fields method."""
        dialogue_message_unspecified = DialogueMessage(
            FipaMessage.Performative.ACCEPT, {}
        )

        is_incoming, target = self._provide_unspecified_fields(
            dialogue_message_unspecified, last_is_incoming=False
        )
        assert is_incoming is True
        assert target is None

        dialogue_message_specified = DialogueMessage(
            FipaMessage.Performative.ACCEPT, {}, False, 4
        )

        is_incoming, target = self._provide_unspecified_fields(
            dialogue_message_specified, last_is_incoming=True
        )
        assert is_incoming is False
        assert target == 4

    def test_non_initial_incoming_message_dialogue_reference(self):
        """Test the _non_initial_incoming_message_dialogue_reference method."""
        dialogue_incomplete_ref = FipaDialogue(
            DialogueLabel(("2", ""), "opponent", "self_address"),
            "self_address",
            FipaDialogue.Role.BUYER,
        )
        reference_incomplete = self._non_initial_incoming_message_dialogue_reference(
            dialogue_incomplete_ref
        )
        assert reference_incomplete[1] != ""

        dialogue_complete_ref = FipaDialogue(
            DialogueLabel(("2", "7"), "opponent", "self_address"),
            "self_address",
            FipaDialogue.Role.BUYER,
        )
        reference_complete = self._non_initial_incoming_message_dialogue_reference(
            dialogue_complete_ref
        )
        assert reference_complete[1] == "7"

    def test_extract_message_fields(self):
        """Test the _extract_message_fields method."""
        expected_performative = FipaMessage.Performative.ACCEPT
        expected_contents = {}
        expected_is_incoming = False
        expected_target = 4
        dialogue_message = DialogueMessage(
            expected_performative,
            expected_contents,
            expected_is_incoming,
            expected_target,
        )

        (
            actual_performative,
            actual_contents,
            actual_message_id,
            actual_is_incoming,
            actual_target,
        ) = self._extract_message_fields(
            message=dialogue_message, index=3, last_is_incoming=True
        )

        assert actual_message_id == 4
        assert actual_target == expected_target
        assert actual_performative == expected_performative
        assert actual_contents == expected_contents
        assert actual_is_incoming == expected_is_incoming

    def test_prepare_skill_dialogue_valid_self_initiated(self):
        """Positive test for prepare_skill_dialogue method with a valid dialogue initiated by self."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": "some_query"}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_1"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_2"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_3"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_4"},
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT, {}),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM, {"info": "some_info"}
            ),
        )
        dialogue = self.prepare_skill_dialogue(
            fipa_dialogues, dialogue_messages, "counterparty",
        )

        assert type(dialogue) == FipaDialogue
        assert dialogue.is_self_initiated
        assert len(dialogue._outgoing_messages) == 4
        assert len(dialogue._incoming_messages) == 4
        assert dialogue._incoming_messages[1].proposal == "some_counter_proposal_2"
        assert dialogue._incoming_messages[3].info == "some_info"

    def test_prepare_skill_dialogue_valid_opponent_initiated(self):
        """Positive test for prepare_skill_dialogue method with a valid dialogue initiated by the opponent."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = (
            DialogueMessage(
                FipaMessage.Performative.CFP, {"query": "some_query"}, True
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_1"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_2"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_3"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_4"},
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT, {}),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM, {"info": "some_info"}
            ),
        )
        dialogue = self.prepare_skill_dialogue(
            fipa_dialogues, dialogue_messages, "counterparty",
        )

        assert type(dialogue) == FipaDialogue
        assert not dialogue.is_self_initiated
        assert len(dialogue._outgoing_messages) == 4
        assert len(dialogue._incoming_messages) == 4
        assert dialogue._outgoing_messages[1].proposal == "some_counter_proposal_2"
        assert dialogue._outgoing_messages[-1].info == "some_info"

    def test_negative_prepare_skill_dialogue_invalid_opponent_initiated(self):
        """Negative test for prepare_skill_dialogue method with an invalid dialogue initiated by the opponent."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = (
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": "some_proposal"}, True
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_1"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_2"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_3"},
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_counter_proposal_4"},
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT, {}),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM, {"info": "some_info"}
            ),
        )
        with pytest.raises(
            AEAEnforceError, match="Cannot update the dialogue with message number 1"
        ):
            self.prepare_skill_dialogue(
                fipa_dialogues, dialogue_messages, "counterparty",
            )

    def test_negative_prepare_skill_dialogue_empty_messages(self):
        """Negative test for prepare_skill_dialogue method where the list of DialogueMessages is emoty."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = tuple()

        with pytest.raises(
            AEAEnforceError, match="the list of messages must be positive."
        ):
            self.prepare_skill_dialogue(
                fipa_dialogues, dialogue_messages, "counterparty",
            )

    def test_negative_prepare_skill_dialogue_invalid(self):
        """Negative test for prepare_skill_dialogue method with an invalid dialogue (a message has invalid target)."""
        fipa_dialogues = FipaDialogues(
            self_address=self.skill.skill_context.agent_address
        )
        dialogue_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": "some_query"}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE,
                {"proposal": "some_proposal"},
                target=2,
            ),
        )

        with pytest.raises(
            AEAEnforceError, match="Cannot update the dialogue with message number .*"
        ):
            self.prepare_skill_dialogue(
                fipa_dialogues, dialogue_messages, "counterparty",
            )


class FipaDialogues(BaseFipaDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address, **kwargs) -> None:
        """
        Initialize dialogues.

        :return: None
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> Dialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return FipaDialogue.Role.BUYER

        BaseFipaDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=FipaDialogue,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

r"""
Wrapper over built-in "open" function.

This module contains a wrapper to the built-in 'open'
function, the 'open_file' function, that fixes the
keyword argument 'newline' to be equal to "\n" (the UNIX line separator).
This will force the line separator to be "\n" both
for incoming and outgoing data.

The reason of this is that files written in an AEA package
need to have "\n" as line separator, on all platforms. Otherwise,
the fingerprint of the packages involved would change across platforms
just because the line separators are replaced.

For instance, the 'open' function on Windows, by default (newline=None),
would replace the line separators "\n" with "\r\n".
This has an impact in the computation of the fingerprint.

Hence, any usage of file system functionalities
should either use 'open_file', or set 'newline="\n"' when
calling the 'open' or the 'pathlib.Path.open' functions.
"""
from functools import partial
from pathlib import Path
from typing import Callable, Optional, TextIO, Union


UNIX_LINESEP = "\n"

_open_file_builtin: Callable = partial(open, newline=UNIX_LINESEP)
_open_file_pathlib: Callable = partial(Path.open, newline=UNIX_LINESEP)

PathNameTypes = Union[int, str, bytes, Path]


def open_file(
    file: PathNameTypes,
    mode: str = "r",
    buffering: int = -1,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
) -> TextIO:
    r"""
    Open a file.

    Behaviour, kwargs and return type are the same for built-in 'open'
    and pathlib.Path.open, except for 'newline', which is fixed to '\n'.

    For more details on the keyword arguments, please refer
    to the documentation for the built-in 'open':

        https://docs.python.org/3/library/functions.html#open

    :param file: either a pathlib.Path object or the type accepted by 'open',
            i.e. a string, bytes or integer.
    :param mode: the mode in which the file is opened.
    :param buffering: the buffering policy.
    :param encoding: the name of the encoding used to decode or encode the file.
    :param errors: how encoding errors are to be handled
    :return: the IO object.
    """
    if "b" in mode:
        raise ValueError("This function can only work in text mode.")
    actual_wrapped_function = _open_file_builtin
    if isinstance(file, Path):
        actual_wrapped_function = _open_file_pathlib
    return actual_wrapped_function(
        file, mode=mode, buffering=buffering, encoding=encoding, errors=errors
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains the behaviours."""

from typing import Any, List, Optional, Set, cast

from aea.protocols.dialogue.base import DialogueLabel
from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.connections.ledger.base import (
    CONNECTION_ID as LEDGER_CONNECTION_PUBLIC_ID,
)
from packages.fetchai.protocols.ledger_api import LedgerApiMessage
from packages.fetchai.skills.generic_buyer.behaviours import GenericSearchBehaviour
from packages.fetchai.skills.ml_train.dialogues import (
    LedgerApiDialogue,
    LedgerApiDialogues,
    MlTradeDialogue,
)


DEFAULT_MAX_PROCESSING = 120
DEFAULT_TX_INTERVAL = 2.0
DEFAULT_SEARCH_INTERVAL = 5.0
LEDGER_API_ADDRESS = str(LEDGER_CONNECTION_PUBLIC_ID)


SearchBehaviour = GenericSearchBehaviour


class TransactionBehaviour(TickerBehaviour):
    """A behaviour to sequentially submit transactions to the blockchain."""

    def __init__(self, **kwargs: Any):
        """Initialize the transaction behaviour."""
        tx_interval = cast(
            float, kwargs.pop("transaction_interval", DEFAULT_TX_INTERVAL)
        )
        self.max_processing = cast(
            float, kwargs.pop("max_processing", DEFAULT_MAX_PROCESSING)
        )
        self.processing_time = 0.0
        self.waiting: List[MlTradeDialogue] = []
        self.processing: Optional[LedgerApiDialogue] = None
        self.timedout: Set[DialogueLabel] = set()
        super().__init__(tick_interval=tx_interval, **kwargs)

    def setup(self) -> None:
        """Setup behaviour."""
        pass

    def act(self) -> None:
        """
        Implement the act.

        :return: None
        """
        if self.processing is not None:
            if self.processing_time <= self.max_processing:
                # already processing
                self.processing_time += self.tick_interval
                return
            self._timeout_processing()
        if len(self.waiting) == 0:
            # nothing to process
            return
        self._start_processing()

    def _start_processing(self) -> None:
        """Process the next transaction."""
        ml_trade_dialogue = self.waiting.pop(0)
        self.context.logger.info(
            f"Processing transaction, {len(self.waiting)} transactions remaining"
        )

        ledger_api_dialogues = cast(
            LedgerApiDialogues, self.context.ledger_api_dialogues
        )
        ledger_api_msg, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=LEDGER_API_ADDRESS,
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
            terms=ml_trade_dialogue.terms,
        )
        ledger_api_dialogue = cast(LedgerApiDialogue, ledger_api_dialogue)
        ledger_api_dialogue.associated_ml_trade_dialogue = ml_trade_dialogue
        self.processing_time = 0.0
        self.processing = ledger_api_dialogue
        self.context.logger.info(
            f"requesting transfer transaction from ledger api for message={ledger_api_msg}..."
        )
        self.context.outbox.put_message(message=ledger_api_msg)

    def teardown(self) -> None:
        """Teardown behaviour."""
        pass

    def _timeout_processing(self) -> None:
        """Timeout processing."""
        if self.processing is None:
            return
        self.timedout.add(self.processing.dialogue_label)
        self.waiting.append(self.processing.associated_ml_trade_dialogue)
        self.processing_time = 0.0
        self.processing = None

    def finish_processing(self, ledger_api_dialogue: LedgerApiDialogue) -> None:
        """
        Finish processing.

        :param ledger_api_dialogue: the ledger api dialogue
        """
        if self.processing == ledger_api_dialogue:
            self.processing_time = 0.0
            self.processing = None
            return
        if ledger_api_dialogue.dialogue_label not in self.timedout:
            raise ValueError(
                f"Non-matching dialogues in transaction behaviour: {self.processing} and {ledger_api_dialogue}"
            )
        self.timedout.remove(ledger_api_dialogue.dialogue_label)
        self.context.logger.debug(
            f"Timeout dialogue in transaction processing: {ledger_api_dialogue}"
        )
        # don't reset, as another might be processing

    def failed_processing(self, ledger_api_dialogue: LedgerApiDialogue) -> None:
        """
        Failed processing.

        Currently, we retry processing indefinitely.

        :param ledger_api_dialogue: the ledger api dialogue
        """
        self.finish_processing(ledger_api_dialogue)
        self.waiting.append(ledger_api_dialogue.associated_ml_trade_dialogue)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Implementation of the parser for configuration file."""
from pathlib import Path
from typing import Dict, Generic, List, TextIO, Type, TypeVar, Union, cast

import yaml

from aea.configurations.base import (
    AgentConfig,
    ComponentConfiguration,
    ComponentId,
    ComponentType,
    ConnectionConfig,
    ContractConfig,
    PACKAGE_TYPE_TO_CONFIG_CLASS,
    PackageConfiguration,
    PackageType,
    ProtocolConfig,
    ProtocolSpecification,
    SkillConfig,
)
from aea.configurations.validation import ConfigValidator, make_jsonschema_base_uri
from aea.helpers.io import open_file
from aea.helpers.yaml_utils import yaml_dump, yaml_dump_all, yaml_load, yaml_load_all


_STARTING_INDEX_CUSTOM_CONFIGS = 1


_ = make_jsonschema_base_uri  # for tests compatibility

T = TypeVar(
    "T",
    AgentConfig,
    SkillConfig,
    ConnectionConfig,
    ContractConfig,
    ProtocolConfig,
    ProtocolSpecification,
)


class BaseConfigLoader:
    """Base class for configuration loader classes."""

    def __init__(self, schema_filename: str) -> None:
        """
        Initialize the parser for configuration files.

        :param schema_filename: the path to the JSON-schema file in 'aea/configurations/schemas'.
        """
        self._validator = ConfigValidator(schema_filename)

    @property
    def validator(self) -> ConfigValidator:
        """Get the json schema validator."""
        return self._validator

    def validate(self, json_data: Dict) -> None:
        """
        Validate a JSON object.

        :param json_data: the JSON data.
        :return: None.
        """
        self.validator.validate(json_data)

    @property
    def required_fields(self) -> List[str]:
        """
        Get the required fields.

        :return: list of required fields.
        """
        return self.validator.required_fields


class ConfigLoader(Generic[T], BaseConfigLoader):
    """Parsing, serialization and validation for package configuration files."""

    def __init__(self, schema_filename: str, configuration_class: Type[T]) -> None:
        """
        Initialize the parser for configuration files.

        :param schema_filename: the path to the JSON-schema file in 'aea/configurations/schemas'.
        :param configuration_class: the configuration class (e.g. AgentConfig, SkillConfig etc.)
        """
        super().__init__(schema_filename)
        self._configuration_class = configuration_class  # type: Type[T]

    @property
    def configuration_class(self) -> Type[T]:
        """Get the configuration class of the loader."""
        return self._configuration_class

    def load_protocol_specification(
        self, file_pointer: TextIO
    ) -> ProtocolSpecification:
        """
        Load an agent configuration file.

        :param file_pointer: the file pointer to the configuration file
        :return: the configuration object.
        :raises
        """
        yaml_data = yaml.safe_load_all(file_pointer)
        yaml_documents = list(yaml_data)
        configuration_file_json = yaml_documents[0]
        if len(yaml_documents) == 1:
            protobuf_snippets_json = {}
            dialogue_configuration = {}  # type: Dict
        elif len(yaml_documents) == 2:
            protobuf_snippets_json = (
                {} if "initiation" in yaml_documents[1] else yaml_documents[1]
            )
            dialogue_configuration = (
                yaml_documents[1] if "initiation" in yaml_documents[1] else {}
            )
        elif len(yaml_documents) == 3:
            protobuf_snippets_json = yaml_documents[1]
            dialogue_configuration = yaml_documents[2]
        else:
            raise ValueError(
                "Incorrect number of Yaml documents in the protocol specification."
            )

        self.validate(configuration_file_json)

        protocol_specification = cast(
            ProtocolSpecification,
            self.configuration_class.from_json(configuration_file_json),
        )
        protocol_specification.protobuf_snippets = protobuf_snippets_json
        protocol_specification.dialogue_config = dialogue_configuration
        return protocol_specification

    def load(self, file_pointer: TextIO) -> T:
        """
        Load a configuration file.

        :param file_pointer: the file pointer to the configuration file
        :return: the configuration object.
        """
        if self.configuration_class.package_type == PackageType.AGENT:
            return cast(T, self._load_agent_config(file_pointer))
        return self._load_component_config(file_pointer)

    def dump(self, configuration: T, file_pointer: TextIO) -> None:
        """Dump a configuration.

        :param configuration: the configuration to be dumped.
        :param file_pointer: the file pointer to the configuration file
        :return: None
        """
        if self.configuration_class.package_type == PackageType.AGENT:
            self._dump_agent_config(cast(AgentConfig, configuration), file_pointer)
        else:
            self._dump_component_config(configuration, file_pointer)

    @classmethod
    def from_configuration_type(
        cls, configuration_type: Union[PackageType, str]
    ) -> "ConfigLoader":
        """Get the configuration loader from the type."""
        configuration_type = PackageType(configuration_type)
        return ConfigLoaders.from_package_type(configuration_type)

    def _load_component_config(self, file_pointer: TextIO) -> T:
        """Load a component configuration."""
        configuration_file_json = yaml_load(file_pointer)
        return self._load_from_json(configuration_file_json)

    def _load_from_json(self, configuration_file_json: Dict) -> T:
        """Load component configuration from JSON object."""
        self.validate(configuration_file_json)
        key_order = list(configuration_file_json.keys())
        configuration_obj = cast(
            T, self.configuration_class.from_json(configuration_file_json)
        )
        configuration_obj._key_order = key_order  # pylint: disable=protected-access
        return configuration_obj

    def load_agent_config_from_json(
        self, configuration_json: List[Dict], validate: bool = True
    ) -> AgentConfig:
        """
        Load agent configuration from configuration json data.

        :param configuration_json: list of dicts with aea configuration

        :return: AgentConfig instance
        """
        if len(configuration_json) == 0:
            raise ValueError("Agent configuration file was empty.")
        agent_config_json = configuration_json[0]
        if validate:
            self.validate(agent_config_json)
        key_order = list(agent_config_json.keys())
        agent_configuration_obj = cast(
            AgentConfig, self.configuration_class.from_json(agent_config_json)
        )
        agent_configuration_obj._key_order = (  # pylint: disable=protected-access
            key_order
        )

        component_configurations = self._get_component_configurations(
            configuration_json
        )
        agent_configuration_obj.component_configurations = component_configurations
        return agent_configuration_obj

    def _get_component_configurations(
        self, configuration_file_jsons: List[Dict]
    ) -> Dict[ComponentId, Dict]:
        """
        Get the component configurations from the tail pages of the aea-config.yaml file.

        :param configuration_file_jsons: the JSON objects of the custom configurations of a aea-config.yaml file.
        :return: a dictionary whose keys are component ids and values are the configurations.
        """
        component_configurations: Dict[ComponentId, Dict] = {}
        # load the other components.
        for i, component_configuration_json in enumerate(
            configuration_file_jsons[_STARTING_INDEX_CUSTOM_CONFIGS:]
        ):
            component_id = self._process_component_section(
                i, component_configuration_json
            )
            if component_id in component_configurations:
                raise ValueError(
                    f"Configuration of component {component_id} occurs more than once."
                )
            component_configurations[component_id] = component_configuration_json
        return component_configurations

    def _load_agent_config(self, file_pointer: TextIO) -> AgentConfig:
        """Load an agent configuration."""
        configuration_file_jsons = yaml_load_all(file_pointer)
        return self.load_agent_config_from_json(configuration_file_jsons)

    def _dump_agent_config(
        self, configuration: AgentConfig, file_pointer: TextIO
    ) -> None:
        """Dump agent configuration."""
        agent_config_part = configuration.ordered_json
        self.validate(agent_config_part)
        agent_config_part.pop("component_configurations")
        result = [agent_config_part] + configuration.component_configurations_json()
        yaml_dump_all(result, file_pointer)

    def _dump_component_config(self, configuration: T, file_pointer: TextIO) -> None:
        """Dump component configuration."""
        result = configuration.ordered_json
        self.validate(result)
        yaml_dump(result, file_pointer)

    def _process_component_section(
        self, component_index: int, component_configuration_json: Dict
    ) -> ComponentId:
        """
        Process a component configuration in an agent configuration file.

        It breaks down in:
        - extract the component id
        - validate the component configuration
        - check that there are only configurable fields

        :param component_index: the index of the component in the file.
        :param component_configuration_json: the JSON object.
        :return: the processed component configuration.
        """
        component_id = self.validator.split_component_id_and_config(
            component_index, component_configuration_json
        )
        self.validator.validate_component_configuration(
            component_id, component_configuration_json
        )
        return component_id


class ConfigLoaders:
    """Configuration Loader class to load any package type."""

    @classmethod
    def from_package_type(
        cls, configuration_type: Union[PackageType, str]
    ) -> "ConfigLoader":
        """
        Get a config loader from the configuration type.

        :param configuration_type: the configuration type
        """
        config_class: Type[PackageConfiguration] = PACKAGE_TYPE_TO_CONFIG_CLASS[
            PackageType(configuration_type)
        ]
        return ConfigLoader(config_class.schema, config_class)


def load_component_configuration(
    component_type: ComponentType,
    directory: Path,
    skip_consistency_check: bool = False,
) -> ComponentConfiguration:
    """
    Load configuration and check that it is consistent against the directory.

    :param component_type: the component type.
    :param directory: the root of the package
    :param skip_consistency_check: if True, the consistency check are skipped.
    :return: the configuration object.
    """
    package_type = component_type.to_package_type()
    configuration_object = load_package_configuration(
        package_type, directory, skip_consistency_check
    )
    configuration_object = cast(ComponentConfiguration, configuration_object)
    return configuration_object


def load_package_configuration(
    package_type: PackageType, directory: Path, skip_consistency_check: bool = False,
) -> PackageConfiguration:
    """
    Load configuration and check that it is consistent against the directory.

    :param package_type: the package type.
    :param directory: the root of the package
    :param skip_consistency_check: if True, the consistency check are skipped.
    :return: the configuration object.
    """
    configuration_object = _load_configuration_object(package_type, directory)
    if not skip_consistency_check and isinstance(
        configuration_object, ComponentConfiguration
    ):
        configuration_object._check_configuration_consistency(  # pylint: disable=protected-access
            directory
        )
    return configuration_object


def _load_configuration_object(
    package_type: PackageType, directory: Path
) -> PackageConfiguration:
    """
    Load the configuration object, without consistency checks.

    :param package_type: the package type.
    :param directory: the directory of the configuration.
    :return: the configuration object.
    :raises FileNotFoundError: if the configuration file is not found.
    """
    configuration_loader = ConfigLoader.from_configuration_type(package_type)
    configuration_filename = (
        configuration_loader.configuration_class.default_configuration_filename
    )
    configuration_filepath = directory / configuration_filename
    try:
        with open_file(configuration_filepath) as fp:
            configuration_object = configuration_loader.load(fp)
    except FileNotFoundError:
        raise FileNotFoundError(
            "{} configuration not found: {}".format(
                package_type.value.capitalize(), configuration_filepath
            )
        )
    return configuration_object



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the helper module."""
import pytest

from aea.helpers.env_vars import (
    apply_env_variables,
    convert_value_str_to_type,
    is_env_variable,
    replace_with_env_var,
)


def test_is_env_variable():
    """Test is_env_variable."""
    assert is_env_variable("${test}")
    assert is_env_variable("${test:int}")
    assert is_env_variable("${test:int:12}")

    assert not is_env_variable("sdfsdf")


def test_apply_env_variables():
    """Test apply_env_variables"""
    assert apply_env_variables("${var}", {"var": "test"}) == "test"
    assert apply_env_variables("var", {"var": "test"}) == "var"
    assert apply_env_variables(["${var}"], {"var": "test"}) == ["test"]
    assert apply_env_variables({"${var}": "${var}"}, {"var": "test"}) == {
        "test": "test"
    }


def test_replace_with_env_var():
    """Test replace_with_env_var."""
    assert replace_with_env_var("${var:int:12}", {"var": "10"}) == 10
    assert replace_with_env_var("${var:int:12}", {}) == 12
    assert replace_with_env_var("var", {}) == "var"
    assert replace_with_env_var("${var}", {}, default_value=100) == 100

    with pytest.raises(
        ValueError, match=r"`var` not found in env variables and no default value set!",
    ):
        replace_with_env_var("${var}", {})


def test_convert_value_str_to_type():
    """Test convert_value_str_to_type."""
    assert convert_value_str_to_type("false", "bool") is False
    assert convert_value_str_to_type("True", "bool") is True
    assert convert_value_str_to_type("12", "int") == 12
    assert convert_value_str_to_type("1.1", "float") == 1.1
    assert convert_value_str_to_type("1sdfsdf2", "none") is None
    assert convert_value_str_to_type('{"a": 12}', "dict") == {"a": 12}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains Negative tests for Libp2p connection."""
import os
import shutil
import tempfile
from asyncio.futures import Future
from unittest.mock import Mock, patch

import pytest

from aea.configurations.base import ConnectionConfig
from aea.crypto.registries import make_crypto
from aea.identity.base import Identity
from aea.multiplexer import Multiplexer

from packages.fetchai.connections.p2p_libp2p.connection import (
    LIBP2P_NODE_MODULE_NAME,
    Libp2pNode,
    P2PLibp2pConnection,
    _golang_module_run,
    _ip_all_private_or_all_public,
)

from tests.conftest import DEFAULT_LEDGER, _make_libp2p_connection


DEFAULT_PORT = 10234
DEFAULT_NET_SIZE = 4


class TestP2PLibp2pConnectionFailureGolangRun:
    """Test that golang run fails if wrong path or timeout"""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        temp_dir = os.path.join(cls.t, "temp_dir")
        os.mkdir(temp_dir)
        cls.connection = _make_libp2p_connection(data_dir=temp_dir)
        cls.wrong_path = tempfile.mkdtemp()

    def test_wrong_path(self):
        """Test the wrong path."""
        log_file_desc = open("log", "a", 1)
        with pytest.raises(Exception):
            _golang_module_run(
                self.wrong_path, LIBP2P_NODE_MODULE_NAME, [], log_file_desc
            )

    def test_timeout(self):
        """Test the timeout."""
        self.connection.node._connection_timeout = 0
        muxer = Multiplexer([self.connection])
        with pytest.raises(Exception):
            muxer.connect()
        muxer.disconnect()

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
            shutil.rmtree(cls.wrong_path)
        except (OSError, IOError):
            pass


class TestP2PLibp2pConnectionFailureNodeDisconnect:
    """Test that connection handles node disconnecting properly"""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)

        temp_dir = os.path.join(cls.t, "temp_dir")
        os.mkdir(temp_dir)
        cls.connection = _make_libp2p_connection(data_dir=temp_dir)

    def test_node_disconnect(self):
        """Test node disconnect."""
        muxer = Multiplexer([self.connection])
        muxer.connect()
        self.connection.node.proc.terminate()
        self.connection.node.proc.wait()
        muxer.disconnect()

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestP2PLibp2pConnectionFailureSetupNewConnection:
    """Test that connection constructor ensures proper configuration"""

    @classmethod
    def setup_class(cls):
        """Set the test up"""
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        os.chdir(cls.t)
        crypto = make_crypto(DEFAULT_LEDGER)
        cls.identity = Identity("", address=crypto.address)
        cls.host = "localhost"
        cls.port = "10000"

        cls.key_file = os.path.join(cls.t, "keyfile")
        key_file_desc = open(cls.key_file, "ab")
        crypto.dump(key_file_desc)
        key_file_desc.close()

    def test_entry_peers_when_no_public_uri_provided(self):
        """Test entry peers when no public uri provided."""
        configuration = ConnectionConfig(
            libp2p_key_file=None,
            local_uri="{}:{}".format(self.host, self.port),
            delegate_uri="{}:{}".format(self.host, self.port),
            entry_peers=None,
            log_file=None,
            connection_id=P2PLibp2pConnection.connection_id,
            build_directory=self.t,
        )
        with pytest.raises(ValueError):
            P2PLibp2pConnection(
                configuration=configuration, data_dir=self.t, identity=self.identity
            )

    def test_local_uri_provided_when_public_uri_provided(self):
        """Test local uri provided when public uri provided."""
        configuration = ConnectionConfig(
            node_key_file=self.key_file,
            public_uri="{}:{}".format(self.host, self.port),
            entry_peers=None,
            log_file=None,
            connection_id=P2PLibp2pConnection.connection_id,
            build_directory=self.t,
        )
        with pytest.raises(ValueError):
            P2PLibp2pConnection(
                configuration=configuration, data_dir=self.t, identity=self.identity
            )

    @classmethod
    def teardown_class(cls):
        """Tear down the test"""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


def test_libp2pconnection_mixed_ip_address():
    """Test correct public uri ip and entry peers ips configuration."""
    assert _ip_all_private_or_all_public([]) is True
    assert _ip_all_private_or_all_public(["127.0.0.1", "127.0.0.1"]) is True
    assert _ip_all_private_or_all_public(["localhost", "127.0.0.1"]) is True
    assert _ip_all_private_or_all_public(["10.0.0.1", "127.0.0.1"]) is False
    assert _ip_all_private_or_all_public(["fetch.ai", "127.0.0.1"]) is False
    assert _ip_all_private_or_all_public(["104.26.2.97", "127.0.0.1"]) is False
    assert _ip_all_private_or_all_public(["fetch.ai", "acn.fetch.ai"]) is True


def test_libp2pconnection_node_config_registration_delay():
    """Test node registration delay configuration"""
    host = "localhost"
    port = "10000"

    with tempfile.TemporaryDirectory() as data_dir:
        _make_libp2p_connection(
            port=port, host=host, data_dir=data_dir, build_directory=data_dir
        )
    with tempfile.TemporaryDirectory() as data_dir:
        with pytest.raises(ValueError):
            _make_libp2p_connection(
                port=port,
                host=host,
                data_dir=data_dir,
                peer_registration_delay="must_be_float",
                build_directory=data_dir,
            )


def test_build_dir_not_set():
    """Test build dir not set."""
    host = "localhost"
    port = "10000"
    with tempfile.TemporaryDirectory() as data_dir:
        con = _make_libp2p_connection(
            port=port, host=host, data_dir=data_dir, build_directory=data_dir
        )
        con.configuration.build_directory = None
        with pytest.raises(
            ValueError, match="Connection Configuration build directory is not set!"
        ):
            P2PLibp2pConnection(
                configuration=con.configuration,
                data_dir=data_dir,
                identity=con._identity,
                crypto_store=con.crypto_store,
            )


@pytest.mark.asyncio
async def test_reconnect_on_write_failed():
    """Test node restart on write fail."""
    node = Libp2pNode(Mock(), Mock(), "tmp", "tmp")
    node.pipe = Mock()
    node.pipe.write = Mock(side_effect=Exception("expected"))
    f = Future()
    f.set_result(None)
    with patch.object(node, "restart", return_value=f) as restart_mock, pytest.raises(
        Exception, match="expected"
    ):
        await node.write(b"some_data")

    assert node.pipe.write.call_count == 2
    restart_mock.assert_called_once()


@pytest.mark.asyncio
async def test_reconnect_on_read_failed():
    """Test node restart on read fail."""
    node = Libp2pNode(Mock(), Mock(), "tmp", "tmp")
    node.pipe = Mock()
    node.pipe.read = Mock(side_effect=Exception("expected"))
    f = Future()
    f.set_result(None)
    with patch.object(node, "restart", return_value=f) as restart_mock:
        result = await node.read()

    assert result is None

    assert node.pipe.read.call_count == 2
    restart_mock.assert_called_once()


@pytest.mark.asyncio
async def test_max_restarts():
    """Test node max restarts exception."""
    node = Libp2pNode(Mock(), Mock(), "tmp", "tmp", max_restarts=0)
    with pytest.raises(ValueError, match="Max restarts attempts reached:"):
        await node.restart()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the dialogue classes of the tac control contract skill."""

from pathlib import Path
from typing import cast

import pytest

from aea.exceptions import AEAEnforceError
from aea.helpers.search.models import (
    Attribute,
    Constraint,
    ConstraintType,
    DataModel,
    Description,
    Query,
)
from aea.helpers.transaction.base import RawTransaction, Terms
from aea.protocols.dialogue.base import DialogueLabel
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_AGENT_ADDRESS

from packages.fetchai.protocols.contract_api.custom_types import Kwargs
from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.protocols.signing.message import SigningMessage
from packages.fetchai.skills.tac_negotiation.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
    DefaultDialogue,
    DefaultDialogues,
    FipaDialogue,
    FipaDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    OefSearchDialogue,
    OefSearchDialogues,
    SigningDialogue,
    SigningDialogues,
)
from packages.fetchai.skills.tac_negotiation.helpers import SUPPLY_DATAMODEL_NAME

from tests.conftest import ROOT_DIR


class TestDialogues(BaseSkillTestCase):
    """Test dialogue classes of tac negotiation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.default_dialogues = cast(
            DefaultDialogues, cls._skill.skill_context.default_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )
        cls.oef_search_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )
        cls.query = Query(
            [Constraint("some_attribute", ConstraintType("==", "some_service"))],
            DataModel(
                SUPPLY_DATAMODEL_NAME,
                [
                    Attribute(
                        "some_attribute", str, False, "Some attribute descriptions."
                    )
                ],
            ),
        )

    def test_fipa_dialogue(self):
        """Test the FipaDialogue class."""
        fipa_dialogue = FipaDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=FipaDialogue.Role.BUYER,
        )

        # counterparty_signature
        with pytest.raises(ValueError, match="counterparty_signature not set!"):
            assert fipa_dialogue.counterparty_signature
        fipa_dialogue.counterparty_signature = "some_counterparty_signature"
        with pytest.raises(
            AEAEnforceError, match="counterparty_signature already set!"
        ):
            fipa_dialogue.counterparty_signature = "some_other_counterparty_signature"
        assert fipa_dialogue.counterparty_signature == "some_counterparty_signature"

        # proposal
        with pytest.raises(ValueError, match="Proposal not set!"):
            assert fipa_dialogue.proposal
        description = Description({"foo1": 1, "bar1": 2})
        fipa_dialogue.proposal = description
        with pytest.raises(AEAEnforceError, match="Proposal already set!"):
            fipa_dialogue.proposal = description
        assert fipa_dialogue.proposal == description

        # terms
        with pytest.raises(ValueError, match="Terms not set!"):
            assert fipa_dialogue.terms
        terms = Terms(
            "some_ledger_id",
            self.skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        fipa_dialogue.terms = terms
        with pytest.raises(AEAEnforceError, match="Terms already set!"):
            fipa_dialogue.terms = terms
        assert fipa_dialogue.terms == terms

    def test_fipa_dialogues(self):
        """Test the FipaDialogues class."""
        _, dialogue = self.fipa_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=FipaMessage.Performative.CFP,
            query=self.query,
        )
        assert dialogue.role == FipaDialogue.Role.BUYER
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_contract_api_dialogue(self):
        """Test the ContractApiDialogue class."""
        contract_api_dialogue = ContractApiDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # associated_fipa_dialogue
        with pytest.raises(ValueError, match="associated_fipa_dialogue not set!"):
            assert contract_api_dialogue.associated_fipa_dialogue

        fipa_dialogue = FipaDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=FipaDialogue.Role.BUYER,
        )
        contract_api_dialogue.associated_fipa_dialogue = fipa_dialogue
        with pytest.raises(
            AEAEnforceError, match="associated_fipa_dialogue already set!"
        ):
            contract_api_dialogue.associated_fipa_dialogue = fipa_dialogue
        assert contract_api_dialogue.associated_fipa_dialogue == fipa_dialogue

    def test_contract_api_dialogues(self):
        """Test the ContractApiDialogues class."""
        _, dialogue = self.contract_api_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            ledger_id="some_ledger_id",
            contract_id="some_contract_id",
            callable="some_callable",
            kwargs=Kwargs({"some_key": "some_value"}),
        )
        assert dialogue.role == ContractApiDialogue.Role.AGENT
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)

    def test_default_dialogues(self):
        """Test the DefaultDialogues class."""
        _, dialogue = self.default_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=DefaultMessage.Performative.BYTES,
            content=b"some_content",
        )
        assert dialogue.role == DefaultDialogue.Role.AGENT
        assert dialogue.self_address == self.skill.skill_context.agent_address

    def test_ledger_api_dialogue(self):
        """Test the LedgerApiDialogue class."""
        ledger_api_dialogue = LedgerApiDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # associated_signing_dialogue
        with pytest.raises(ValueError, match="Associated signing dialogue not set!"):
            assert ledger_api_dialogue.associated_signing_dialogue
        signing_dialogue = SigningDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=SigningDialogue.Role.SKILL,
        )
        ledger_api_dialogue.associated_signing_dialogue = signing_dialogue
        with pytest.raises(
            AEAEnforceError, match="Associated signing dialogue already set!"
        ):
            ledger_api_dialogue.associated_signing_dialogue = signing_dialogue
        assert ledger_api_dialogue.associated_signing_dialogue == signing_dialogue

    def test_ledger_api_dialogues(self):
        """Test the LedgerApiDialogues class."""
        _, dialogue = self.ledger_api_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id="some_ledger_id",
            address="some_address",
        )
        assert dialogue.role == LedgerApiDialogue.Role.AGENT
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)

    def test_oef_search_dialogue(self):
        """Test the OefSearchDialogue class."""
        oef_search_dialogue = OefSearchDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # is_seller_search
        with pytest.raises(ValueError, match="is_seller_search not set!"):
            assert oef_search_dialogue.is_seller_search
        oef_search_dialogue.is_seller_search = True
        with pytest.raises(AEAEnforceError, match="is_seller_search already set!"):
            oef_search_dialogue.is_seller_search = False
        assert oef_search_dialogue.is_seller_search is True

    def test_oef_search_dialogues(self):
        """Test the OefSearchDialogues class."""
        _, dialogue = self.oef_search_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query=self.query,
        )
        assert dialogue.role == OefSearchDialogue.Role.AGENT
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)

    def test_signing_dialogue(self):
        """Test the SigningDialogue class."""
        signing_dialogue = SigningDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=ContractApiDialogue.Role.AGENT,
        )

        # associated_contract_api_dialogue
        with pytest.raises(ValueError, match="associated_fipa_dialogue not set!"):
            assert signing_dialogue.associated_fipa_dialogue
        fipa_dialogue = FipaDialogue(
            DialogueLabel(
                ("", ""),
                COUNTERPARTY_AGENT_ADDRESS,
                self.skill.skill_context.agent_address,
            ),
            self.skill.skill_context.agent_address,
            role=FipaDialogue.Role.BUYER,
        )
        signing_dialogue.associated_fipa_dialogue = fipa_dialogue
        with pytest.raises(
            AEAEnforceError, match="associated_fipa_dialogue already set!"
        ):
            signing_dialogue.associated_fipa_dialogue = fipa_dialogue
        assert signing_dialogue.associated_fipa_dialogue == fipa_dialogue

    def test_signing_dialogues(self):
        """Test the SigningDialogues class."""
        _, dialogue = self.signing_dialogues.create(
            counterparty=COUNTERPARTY_AGENT_ADDRESS,
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            terms=Terms(
                "some_ledger_id",
                "some_sender_address",
                "some_counterparty_address",
                dict(),
                dict(),
                "some_nonce",
            ),
            raw_transaction=RawTransaction(
                "some_ledger_id", {"some_key": "some_value"}
            ),
        )
        assert dialogue.role == SigningDialogue.Role.SKILL
        assert dialogue.self_address == str(self.skill.skill_context.skill_id)
